"""Collect information about the environment and display it.

- Python version
- Package availablility and versions: Zeus, PyTorch, JAX.
- NVIDIA GPU availability: Number of GPUs and moels.
- AMD GPU availability: Number of GPUs and models.
- Intel RAPL availability: Number of CPUs and whether DRAM measurements are available.
"""

from __future__ import annotations

import platform

import zeus
from zeus.utils import framework
from zeus.device import get_gpus, get_cpus
from zeus.device.cpu import RAPLCPUs
from zeus.device.gpu.common import ZeusGPUInitError, EmptyGPUs
from zeus.device.cpu.common import ZeusCPUInitError, EmptyCPUs


SECTION_SEPARATOR = "=" * 80 + "\n"


def show_env():
    """Collect information about the environment and display it."""
    print(SECTION_SEPARATOR)
    print(f"Python version: {platform.python_version()}\n")

    print(SECTION_SEPARATOR)
    package_availability = "\nPackage availability and versions:\n"
    package_availability += f"  Zeus: {zeus.__version__}\n"

    try:
        torch_available = framework.torch_is_available()
        torch_cuda_available = True
    except RuntimeError:
        torch_available = framework.torch_is_available(ensure_cuda=False)
        torch_cuda_available = False

    if torch_available and torch_cuda_available:
        torch = framework.MODULE_CACHE["torch"]
        package_availability += f"  PyTorch: {torch.__version__} (with CUDA support)\n"
    elif torch_available and not torch_cuda_available:
        torch = framework.MODULE_CACHE["torch"]
        package_availability += (
            f"  PyTorch: {torch.__version__} (without CUDA support)\n"
        )
    else:
        package_availability += "  PyTorch: not available\n"

    try:
        jax_available = framework.jax_is_available()
        jax_cuda_available = True
    except RuntimeError:
        jax_available = framework.jax_is_available(ensure_cuda=False)
        jax_cuda_available = False

    if jax_available and jax_cuda_available:
        jax = framework.MODULE_CACHE["jax"]
        package_availability += f"  JAX: {jax.__version__} (with CUDA support)\n"
    elif jax_available and not jax_cuda_available:
        jax = framework.MODULE_CACHE["jax"]
        package_availability += f"  JAX: {jax.__version__} (without CUDA support)\n"
    else:
        package_availability += "  JAX: not available\n"

    print(package_availability)

    print(SECTION_SEPARATOR)
    gpu_availability = "\nGPU availability:\n"
    try:
        gpus = get_gpus()
    except ZeusGPUInitError:
        gpus = EmptyGPUs()
    if len(gpus) > 0:
        for i in range(len(gpus)):
            gpu_availability += f"  GPU {i}: {gpus.getName(i)}\n"
    else:
        gpu_availability += "  No GPUs available.\n"
    print(gpu_availability)

    print(SECTION_SEPARATOR)
    cpu_availability = "\nCPU availability:\n"
    try:
        cpus = get_cpus()
    except ZeusCPUInitError:
        cpus = EmptyCPUs()
    if len(cpus) > 0:
        assert isinstance(cpus, RAPLCPUs)
        for i in range(len(cpus)):
            cpu_availability += f"  CPU {i}:\n    CPU measurements available ({cpus.cpus[i].rapl_file.path})\n"
            if cpus.supportsGetDramEnergyConsumption(i):
                dram = cpus.cpus[i].dram
                assert dram is not None
                cpu_availability += f"    DRAM measurements available ({dram.path})\n"
            else:
                cpu_availability += "    DRAM measurements unavailable\n"
    else:
        cpu_availability += "  No CPUs available.\n"
    print(cpu_availability)

    print(SECTION_SEPARATOR)


if __name__ == "__main__":
    show_env()



from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Literal

import pandas as pd
import pytest
from pytest_mock import MockerFixture
from tests.optimizer.batch_size.simulate_with_server import SimulatorWithServer
from zeus._legacy.policy.optimizer import (
    JITPowerLimitOptimizer,
    PruningGTSBatchSizeOptimizer,
)
from zeus._legacy.simulate import Simulator
from zeus._legacy.job import Job

config = {
    "gpu": "v100",
    "eta_knob": 0.5,
    "beta_knob": 2.0,
    "seed": 1,
    "dataset": "librispeech",
    "model": "deepspeech2",
    "optimizer": "adamw",
    "target_metric": 40.0,
    "max_epochs": 16,
    "b_0": 192,  # default_bs
    "num_recurrence": None,
}


def arm_state_parser(output):
    # Define regex patterns to match the numbers
    arm_pattern = r"Arm\s+(\d+)"
    mu_pattern = r"N\(([-+]?\d*\.\d+|\d+),\s+([-+]?\d*\.\d+|\d+)\)"
    arrow_pattern = r"-> ([-+]?\d*\.\d+|\d+)"

    # Use regex to find the numbers in the output
    arm_numbers = re.findall(arm_pattern, output, re.MULTILINE)
    mu_numbers = re.findall(mu_pattern, output, re.MULTILINE)
    arrow_numbers = re.findall(arrow_pattern, output, re.MULTILINE)

    d = []

    for arm, mu, arrow in zip(arm_numbers, mu_numbers, arrow_numbers):
        d.append({"Arm": arm, "Mean": mu[0], "stdev": mu[1], "Arrow": arrow})
    d.sort(key=lambda x: x["Arm"])
    return d


def read_trace(
    gpu: Literal["a40", "v100", "p100", "rtx6000"]
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Read the train and power trace files as Pandas DataFrames."""
    trace_dir = Path(__file__).resolve(strict=True).parents[3]
    train_df = pd.DataFrame(pd.read_csv(trace_dir / "trace/summary_train.csv"))
    power_df = pd.DataFrame(pd.read_csv(trace_dir / f"trace/summary_power_{gpu}.csv"))
    return train_df, power_df


@pytest.fixture(scope="module", autouse=True)
def logger_setup():
    logger = logging.getLogger(
        "zeus.optimizer.batch_size.server.mab"
    )  # for testing, propagate the log to the root logger so that caplog can capture
    logger.propagate = True


def test_end_to_end(client, caplog, capsys, mocker: MockerFixture):
    mocker.patch("httpx.post", side_effect=client.post)
    mocker.patch("httpx.get", side_effect=client.get)
    mocker.patch("httpx.patch", side_effect=client.patch)

    ## CONFIG
    gpu: Literal["a40", "v100", "p100", "rtx6000"] = config["gpu"]
    eta_knob: float = config["eta_knob"]
    beta_knob: float = config["beta_knob"]
    num_recurrence: int | None = config["num_recurrence"]
    seed: int = config["seed"]

    job = Job(
        config["dataset"],
        config["model"],
        config["optimizer"],
        config["target_metric"],
        config["max_epochs"],
        config["b_0"],
    )

    train_df, power_df = read_trace(gpu)

    # # Use 2 * |B| * |P| is num_recurrence is None.
    # print(num_recurrence)
    if num_recurrence is None:
        job_df = job.filter_df(train_df.merge(power_df, how="inner"))
        num_recurrence = (
            2 * len(job_df.batch_size.unique()) * len(job_df.power_limit.unique())
        )

    ### New simulator
    # Instantiate optimizers.
    plo = JITPowerLimitOptimizer(verbose=False)

    # Instantitate the simulator.
    simulator = SimulatorWithServer(
        train_df, power_df, plo, gpu, seed=seed, verbose=False
    )
    # # Run the simulator.
    result = simulator.simulate_one_job(job, num_recurrence, beta_knob, eta_knob)
    selected_bs = [item.bs for item in result]

    ### Original Simulator
    # Instantiate optimizers.
    org_plo = JITPowerLimitOptimizer(verbose=False)
    org_bso = PruningGTSBatchSizeOptimizer(seed=seed, verbose=True)

    # Instantitate the simulator.
    original_simulator = Simulator(
        train_df, power_df, org_bso, org_plo, seed=seed, verbose=False
    )
    original_result = original_simulator.simulate_one_job(
        job, num_recurrence, beta_knob, eta_knob
    )
    org_selected_bs = [item.bs for item in original_result]

    out, err = capsys.readouterr()
    records = arm_state_parser(out)

    new_sim_records = arm_state_parser(caplog.text)

    # Compare arm states
    assert records == new_sim_records

    # Compare selected batch sizes
    assert selected_bs == org_selected_bs



"""Optimize the energy consumption of large model training with Perseus.

A a high-level, this optimizer assigns each forward and backward computation
in a pipeline parallel training iteration with a GPU frequency that leads to
a Pareto-optimal training iteration time and energy consumption.

Currently, this optimizer depends on PyTorch.
"""

from zeus.optimizer.pipeline_frequency.optimizer import PipelineFrequencyOptimizer



from __future__ import annotations

from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture
from zeus.monitor.energy import Measurement, ZeusMonitor
from zeus.optimizer.batch_size.client import BatchSizeOptimizer
from zeus.optimizer.batch_size.common import JobSpec
from zeus.optimizer.batch_size.exceptions import (
    ZeusBSOBadOperationError,
    ZeusBSOTrainFailError,
)


@pytest.fixture
def mock_monitor(mocker: MockerFixture):
    mocker.patch("pynvml.nvmlInit")

    zeus_monitor_mock_instance = MagicMock(spec=ZeusMonitor)
    zeus_monitor_mock_instance.gpu_indices = [0, 1, 2, 3]
    zeus_monitor_mock_instance.end_window.return_value = Measurement(
        time=37.24807469360,
        gpu_energy={
            0: 4264.87199999392,
            1: 4367.186999991536,
            2: 4342.869000002742,
            3: 4158.034000009298,
        },
    )

    mocker.patch(
        "zeus.monitor.energy.ZeusMonitor", return_value=zeus_monitor_mock_instance
    )
    mocker.patch("pynvml.nvmlDeviceGetName").return_value = "Tesla V100"
    mocker.patch("pynvml.nvmlDeviceGetCount").return_value = 4
    mocker.patch("pynvml.nvmlDeviceGetHandleByIndex").return_value = 0
    mocker.patch("pynvml.nvmlDeviceGetPowerManagementLimitConstraints").return_value = [
        100000,
        300000,
    ]
    return zeus_monitor_mock_instance


@pytest.fixture(autouse=True)
def mock_http_call(client, mocker: MockerFixture):
    mocker.patch("httpx.post", side_effect=client.post)
    mocker.patch("httpx.get", side_effect=client.get)
    mocker.patch("httpx.patch", side_effect=client.patch)

    mocker.patch("atexit.register")


def test_register_job(mock_monitor, helpers):
    job = JobSpec.parse_obj(helpers.get_fake_job("test_register_job"))
    bso_client = BatchSizeOptimizer(mock_monitor, "", job)
    assert bso_client.job.max_power == 300 * len(mock_monitor.gpu_indices)
    bso_client = BatchSizeOptimizer(mock_monitor, "", job)
    assert bso_client.job.max_power == 300 * len(mock_monitor.gpu_indices)


def test_batch_sizes(mock_monitor, helpers):
    job = JobSpec.parse_obj(helpers.get_fake_job("test_batch_sizes"))
    bso_client = BatchSizeOptimizer(mock_monitor, "", job)
    bs = bso_client.get_batch_size()

    assert bs == 1024 and bso_client.current_batch_size == 1024

    bso_client.on_train_begin()
    bso_client.on_evaluate(0.1)
    bso_client.on_evaluate(0.2)
    bso_client.on_evaluate(0.6)  # Converged

    bso_client = BatchSizeOptimizer(mock_monitor, "", job)
    bs = bso_client.get_batch_size()
    bso_client.on_train_begin()

    assert bs == 512 and bso_client.current_batch_size == 512

    i = 0
    with pytest.raises(Exception) as e_info:
        while i < job.max_epochs - 10:  # Test Early stop
            bso_client.on_evaluate(0.3)
            i += 1
            assert i == bso_client.cur_epoch
            assert bso_client.current_batch_size == 512

        assert str(e_info.value).find("cost upper bound") != -1

    bso_client = BatchSizeOptimizer(mock_monitor, "", job)
    bs = bso_client.get_batch_size()
    bso_client.on_train_begin()

    assert bs == 2048 and bso_client.current_batch_size == 2048


def test_converge_fail(mock_monitor, helpers):
    job = JobSpec.parse_obj(helpers.get_fake_job("test_converge_fail"))
    job.beta_knob = None  # disable early stop
    bso_client = BatchSizeOptimizer(mock_monitor, "", job)
    bso_client.on_train_begin()
    bs = bso_client.get_batch_size()

    assert bs == 1024 and bso_client.current_batch_size == 1024

    i = 0
    with pytest.raises(Exception) as e_info:
        while i < job.max_epochs + 10:  # Fail after max_epoch
            bso_client.on_evaluate(0.3)
            i += 1
            assert i == bso_client.cur_epoch
            assert bso_client.current_batch_size == 1024

        print(e_info.value, i)
        assert str(e_info.value).find("Train failed to converge within max_epoch") != -1

    bso_client = BatchSizeOptimizer(mock_monitor, "", job)
    bs = bso_client.get_batch_size()
    bso_client.on_train_begin()

    assert bs == 2048 and bso_client.current_batch_size == 2048


def test_distributed_setting(mock_monitor, helpers):
    job = JobSpec.parse_obj(helpers.get_fake_job("test_distributed_setting"))
    NGPU = 4
    bso_clients = [
        BatchSizeOptimizer(mock_monitor, "", job, rank=i) for i in range(NGPU)
    ]

    # Only rank=0 can ask for bs
    for i in range(1, NGPU):
        with pytest.raises(ZeusBSOBadOperationError) as e_info:
            bso_clients[i].get_batch_size()

    bs = bso_clients[0].get_batch_size()
    # distribute batch size to other clients
    for i in range(1, NGPU):
        bso_clients[i].current_batch_size = bs
        bso_clients[i].trial_number = bso_clients[0].trial_number

    # Mark as unconverged from rank = 0 client
    i = 0
    with pytest.raises(ZeusBSOTrainFailError) as e_info:
        while i < job.max_epochs + 10:  # Fail after max_epoch
            bso_clients[0].on_evaluate(0.3)
            i += 1
            assert i == bso_clients[0].cur_epoch
            assert bso_clients[0].current_batch_size == 1024

        print("[ERROR]", e_info.value, i)
        assert str(e_info.value).find("Train failed to converge within max_epoch") != -1

    for i in range(1, NGPU):
        with pytest.raises(ZeusBSOTrainFailError) as e_info:
            bso_clients[i].on_evaluate(0.3)
            assert bso_clients[i].current_batch_size == 1024
        assert str(e_info.value).find("is already reported.") != -1



"""Commands to use `JobStateRepository`."""

from __future__ import annotations

import json
from typing import Any, Optional

import numpy as np
from zeus.utils.pydantic_v1 import root_validator, validator, Field, BaseModel
from zeus.optimizer.batch_size.common import GpuConfig, JobSpecFromClient, JobParams
from zeus.optimizer.batch_size.server.database.schema import BatchSizeTable, JobTable
from zeus.optimizer.batch_size.server.job.models import Stage


class UpdateExpDefaultBs(BaseModel):
    """Parameters to update the exploration default batch size.

    Attributes:
        job_id: Job Id.
        exp_default_batch_size: new default batch size to use.
    """

    job_id: str
    exp_default_batch_size: int = Field(gt=0)


class UpdateJobStage(BaseModel):
    """Parameters to update the job stage.

    Attributes:
        job_id: Job Id.
        stage: Set it to MAB since we only go from Pruning to MAB.
    """

    job_id: str
    stage: Stage = Field(Stage.MAB, const=True)


class UpdateGeneratorState(BaseModel):
    """Parameters to update the generator state.

    Attributes:
        job_id: Job Id.
        state: Generator state.
    """

    job_id: str
    state: str

    @validator("state")
    def _validate_state(cls, state: str) -> str:
        """Validate the sanity of state."""
        try:
            np.random.default_rng(1).__setstate__(json.loads(state))
            return state
        except (TypeError, ValueError) as err:
            raise ValueError(f"Invalid generator state ({state})") from err


class UpdateJobMinCost(BaseModel):
    """Parameters to update the min training cost and corresponding batch size.

    Attributes:
        job_id: Job Id.
        min_cost: Min training cost.
        min_cost_batch_size: Corresponding batch size.
    """

    job_id: str
    min_cost: float = Field(ge=0)
    min_cost_batch_size: int = Field(gt=0)


class CreateJob(GpuConfig, JobParams):
    """Parameters to create a new job.

    Attributes:
        exp_default_batch_size: Exploration default batch size that is used during Pruning stage.
        min_cost: Min training cost observed. Initially, None.
        min_cost_batch_size: Batch size that has minimum training cost observed.
        stage: Stage of the job.
        mab_random_generator_state: Generator state if mab_seed is not None. Otherwise, None.

    For the rest of attributes, refer to `JobParams`[zeus.optimizer.batch_size.common.JobParams] and `GpuConfig`[zeus.optimizer.batch_size.common.GpuConfig]
    """

    exp_default_batch_size: int
    min_cost: None = Field(None, const=True)
    min_cost_batch_size: int
    stage: Stage = Field(Stage.Pruning, const=True)
    mab_random_generator_state: Optional[str] = None

    class Config:
        """Model configuration.

        Make it immutable after creation.
        """

        frozen = True

    @root_validator(skip_on_failure=True)
    def _validate_states(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate Job states.

        We are checking,
            - If mab seed and generator state is matching.
            - If default, exp_default, min batch sizes are correctly intialized.
            - If default batch size is in the list of batch sizes.
        """
        state: str | None = values["mab_random_generator_state"]
        mab_seed: int | None = values["mab_seed"]
        bss: list[int] = values["batch_sizes"]
        dbs: int = values["default_batch_size"]
        ebs: int = values["exp_default_batch_size"]
        mbs: int = values["min_cost_batch_size"]

        if mab_seed is not None:
            if state is None:
                raise ValueError("mab_seed is not none, but generator state is none")
            else:
                try:
                    np.random.default_rng(1).__setstate__(json.loads(state))
                except (TypeError, ValueError) as err:
                    raise ValueError(f"Invalid generator state ({state})") from err

        if not (dbs == ebs == mbs):
            raise ValueError(
                f"During initialization, default_batch_size({dbs}), exp_default_batch_size({ebs}), min_batch_size({mbs}) should be all the same"
            )
        if dbs not in bss:
            raise ValueError(
                f"default_batch_size({dbs}) is not in the batch size list({bss})"
            )

        return values

    @classmethod
    def from_job_config(cls, js: JobSpecFromClient) -> "CreateJob":
        """From JobConfig, instantiate `CreateJob`.

        Initialize generator state, exp_default_batch_size, and min_cost_batch_size.
        """
        d = js.dict()
        d["exp_default_batch_size"] = js.default_batch_size
        if js.mab_seed is not None:
            rng = np.random.default_rng(js.mab_seed)
            d["mab_random_generator_state"] = json.dumps(rng.__getstate__())
        d["min_cost_batch_size"] = js.default_batch_size
        return cls.parse_obj(d)

    def to_orm(self) -> JobTable:
        """Convert pydantic model `CreateJob` to ORM object Job."""
        d = self.dict()
        job = JobTable()
        for k, v in d.items():
            if k != "batch_sizes":
                setattr(job, k, v)
        job.batch_sizes = [
            BatchSizeTable(job_id=self.job_id, batch_size=bs) for bs in self.batch_sizes
        ]
        return job



"""Generate the Capriccio dataset.

This script takes as input the path to the Sentiment140 dataset
and slices it into 38 overlapping slices. Each slice has 425000
tweets for training (%d_train.json) and 75000 tweets for eval
(%d_val.json).

Usage:
    python capriccio/generate.py /path/to/sentiment140.json
"""

import argparse
import os
import warnings
from datetime import datetime, timedelta
from pathlib import Path

import datasets
import numpy as np
import pandas as pd


def main(output_dir: str) -> None:
    """Run the main routine."""
    # Prepare raw dataset
    print("Preparing raw dataset.")
    df = datasets.load_dataset("sentiment140")["train"].to_pandas()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        df["date"] = pd.to_datetime(df.date)
    df["ind"] = df.index
    df = df.set_index("date")
    df = df.sort_index()
    df = df.rename({"sentiment": "label"}, axis=1)
    df = df.drop(["user", "query"], axis=1)

    # Slice datasets
    stride = timedelta(days=1)
    size = 500_000

    print(f"Slicing datasets with stride {stride} and size {size}.")

    slice_index = 1
    sliced = []
    # Skip April since there are too many days with no tweets.
    now = datetime(year=2009, month=5, day=1)
    end = df.index.max()
    while now < end:
        loc = df.index.get_loc(now.strftime("%m/%d/%Y")).start
        slice_df = df.iloc[loc : loc + size]
        if len(slice_df) < size:
            break

        # Compute sample overlap ratio
        if sliced:
            overlap = len(sliced[-1].merge(slice_df, how="inner", on="ind"))
            print(
                f"{slice_index:2d}: {slice_df.index.min()} ~ {slice_df.index.max()}, overlap = {overlap/size:.3f}"
            )
        else:
            print(f"{slice_index:2d}: {slice_df.index.min()} ~ {slice_df.index.max()}")

        sliced.append(slice_df)

        slice_index += 1
        now += stride

    print(f"{len(sliced)} datasets of size {size} were created.")

    # Split train and validation sets and save
    print("Sampling validation set and saving.")
    seed = 42
    train_frac = 0.85
    save_dir = Path(output_dir)
    os.makedirs(save_dir, exist_ok=True)

    for slice_index, dataset in enumerate(sliced):
        ind = np.random.default_rng(seed).permutation(len(dataset))
        shuffled = dataset.iloc[ind]
        boundary = int(len(dataset) * train_frac)
        train = shuffled.iloc[:boundary].drop(["ind"], axis=1).reset_index()
        val = shuffled.iloc[boundary:].drop(["ind"], axis=1).reset_index()
        train.to_json(
            save_dir / f"{slice_index+1}_train.json", orient="records", lines=True
        )
        val.to_json(
            save_dir / f"{slice_index+1}_val.json", orient="records", lines=True
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir", "-o", required=True, help="Directory to save Capriccio"
    )
    args = parser.parse_args()

    main(args.output_dir)



"""
Forked from https://huggingface.co/google/gemma-7b/blob/359f554e6b7dcfe8026a8705a20ff166141a5adf/examples/example_sft_qlora.py
"""

from dataclasses import dataclass, field
from typing import Optional

import torch

from transformers import AutoTokenizer, HfArgumentParser, AutoModelForCausalLM, BitsAndBytesConfig, TrainingArguments
from datasets import load_dataset
from peft import LoraConfig
from trl import SFTTrainer
from zeus.monitor import ZeusMonitor
from zeus.optimizer.power_limit import HFGlobalPowerLimitOptimizer

@dataclass
class ScriptArguments:
    """
    These arguments vary depending on how many GPUs you have, what their capacity and features are, and what size model you want to train.
    """
    per_device_train_batch_size: Optional[int] = field(default=1)
    per_device_eval_batch_size: Optional[int] = field(default=1)
    gradient_accumulation_steps: Optional[int] = field(default=2)
    learning_rate: Optional[float] = field(default=2e-4)
    max_grad_norm: Optional[float] = field(default=0.3)
    weight_decay: Optional[int] = field(default=0.001)
    lora_alpha: Optional[int] = field(default=16)
    lora_dropout: Optional[float] = field(default=0.1)
    lora_r: Optional[int] = field(default=8)
    max_seq_length: Optional[int] = field(default=2048)
    model_name: Optional[str] = field(
        default=None,
        metadata={
            "help": "The model that you want to train from the Hugging Face hub. E.g. gpt2, gpt2-xl, bert, etc."
        }
    )
    dataset_name: Optional[str] = field(
        default="stingning/ultrachat",
        metadata={"help": "The preference dataset to use."},
    )
    fp16: Optional[bool] = field(
        default=False,
        metadata={"help": "Enables fp16 training."},
    )
    bf16: Optional[bool] = field(
        default=False,
        metadata={"help": "Enables bf16 training."},
    )
    packing: Optional[bool] = field(
        default=True,
        metadata={"help": "Use packing dataset creating."},
    )
    gradient_checkpointing: Optional[bool] = field(
        default=True,
        metadata={"help": "Enables gradient checkpointing."},
    )
    use_flash_attention_2: Optional[bool] = field(
        default=False,
        metadata={"help": "Enables Flash Attention 2."},
    )
    optim: Optional[str] = field(
        default="paged_adamw_32bit",
        metadata={"help": "The optimizer to use."},
    )
    lr_scheduler_type: str = field(
        default="constant",
        metadata={"help": "Learning rate schedule. Constant a bit better than cosine, and has advantage for analysis"},
    )
    max_steps: int = field(default=1000, metadata={"help": "How many optimizer update steps to take"})
    warmup_ratio: float = field(default=0.03, metadata={"help": "Fraction of steps to do a warmup for"})
    save_steps: int = field(default=10, metadata={"help": "Save checkpoint every X updates steps."})
    logging_steps: int = field(default=10, metadata={"help": "Log every X updates steps."})
    output_dir: str = field(
        default="./results",
        metadata={"help": "The output directory where the model predictions and checkpoints will be written."},
    )

parser = HfArgumentParser(ScriptArguments)
script_args = parser.parse_args_into_dataclasses()[0]


def formatting_func(example):
    text = f"### USER: {example['data'][0]}\n### ASSISTANT: {example['data'][1]}"
    return text

# Load the GG model - this is the local one, update it to the one on the Hub
model_id = "google/gemma-7b"

quantization_config = BitsAndBytesConfig(
    load_in_4bit=True,
    bnb_4bit_compute_dtype=torch.float16,
    bnb_4bit_quant_type="nf4"
)

# Load model
model = AutoModelForCausalLM.from_pretrained(
    model_id, 
    quantization_config=quantization_config, 
    torch_dtype=torch.float32,
    # attn_implementation="sdpa" if not script_args.use_flash_attention_2 else "flash_attention_2"
)

# Load tokenizer
tokenizer = AutoTokenizer.from_pretrained(model_id)
tokenizer.pad_token_id = tokenizer.eos_token_id

lora_config = LoraConfig(
    r=script_args.lora_r,
    target_modules=["q_proj", "o_proj", "k_proj", "v_proj", "gate_proj", "up_proj", "down_proj"],
    bias="none",
    task_type="CAUSAL_LM",
    lora_alpha=script_args.lora_alpha,
    lora_dropout=script_args.lora_dropout
)

train_dataset = load_dataset(script_args.dataset_name, split="train[:5%]")

output_dir = "gemma-qlora-ultrachat"

training_arguments = TrainingArguments(
    output_dir=output_dir,
    per_device_train_batch_size=script_args.per_device_train_batch_size,
    gradient_accumulation_steps=script_args.gradient_accumulation_steps,
    optim=script_args.optim,
    save_steps=script_args.save_steps,
    logging_steps=script_args.logging_steps,
    learning_rate=script_args.learning_rate,
    max_grad_norm=script_args.max_grad_norm,
    max_steps=script_args.max_steps,
    warmup_ratio=script_args.warmup_ratio,
    lr_scheduler_type=script_args.lr_scheduler_type,
    gradient_checkpointing=script_args.gradient_checkpointing,
    fp16=script_args.fp16,
    bf16=script_args.bf16,
)

monitor = ZeusMonitor()
# Since one iteration is in the order of tens of seconds or even minutes,
# no need to profile for too many iterations.
optimizer = HFGlobalPowerLimitOptimizer(monitor, warmup_steps=1, profile_steps=3)

trainer = SFTTrainer(
    model=model,
    args=training_arguments,
    train_dataset=train_dataset,
    peft_config=lora_config,
    packing=script_args.packing,
    dataset_text_field="id",
    tokenizer=tokenizer,
    max_seq_length=script_args.max_seq_length,
    formatting_func=formatting_func,
    callbacks=[optimizer],
)

trainer.train()



"""Repository for manipulating Job table."""

from __future__ import annotations

from sqlalchemy import select
from sqlalchemy.ext.asyncio.session import AsyncSession
from zeus.optimizer.batch_size.server.database.repository import DatabaseRepository
from zeus.optimizer.batch_size.server.database.schema import JobTable
from zeus.optimizer.batch_size.server.exceptions import (
    ZeusBSOServiceBadOperationError,
    ZeusBSOValueError,
)
from zeus.optimizer.batch_size.server.job.commands import (
    CreateJob,
    UpdateExpDefaultBs,
    UpdateGeneratorState,
    UpdateJobMinCost,
    UpdateJobStage,
)
from zeus.optimizer.batch_size.server.job.models import JobState
from zeus.utils.logging import get_logger

logger = get_logger(__name__)


class JobStateRepository(DatabaseRepository):
    """Repository that provides basic interfaces to interact with Job table."""

    def __init__(self, session: AsyncSession):
        """Set db session and intialize job. We are working with only one job per session."""
        super().__init__(session)
        self.fetched_job: JobTable | None = None

    async def get_job(self, job_id: str) -> JobState | None:
        """Get job State, which includes jobSpec + batch_sizes(list[int]), without specific states of each batch_size.

        Args:
            job_id: Job id.

        Returns:
            set fetched_job and return `JobState` if we found a job, unless return None.
        """
        stmt = select(JobTable).where(JobTable.job_id == job_id)
        job = await self.session.scalar(stmt)

        if job is None:
            logger.info("get_job: NoResultFound")
            return None

        self.fetched_job = job
        return JobState.from_orm(job)

    def get_job_from_session(self, job_id: str) -> JobState | None:
        """Get a job that was fetched from this session.

        Args:
            job_id: Job id.

        Returns:
            Corresponding `JobState`. If none was found, return None.
        """
        if self.fetched_job is None or self.fetched_job.job_id != job_id:
            return None
        return JobState.from_orm(self.fetched_job)

    def update_exp_default_bs(self, updated_bs: UpdateExpDefaultBs) -> None:
        """Update exploration default batch size on fetched job.

        Args:
            updated_bs: Job Id and new batch size.
        """
        if self.fetched_job is None:
            raise ZeusBSOServiceBadOperationError("No job is fetched.")

        if updated_bs.job_id == self.fetched_job.job_id:
            self.fetched_job.exp_default_batch_size = updated_bs.exp_default_batch_size
        else:
            raise ZeusBSOValueError(
                f"Unknown job_id ({updated_bs.job_id}). Expecting {self.fetched_job.job_id}"
            )

    def update_stage(self, updated_stage: UpdateJobStage) -> None:
        """Update stage on fetched job.

        Args:
            updated_stage: Job Id and new stage.
        """
        if self.fetched_job is None:
            raise ZeusBSOServiceBadOperationError("No job is fetched.")

        if self.fetched_job.job_id == updated_stage.job_id:
            self.fetched_job.stage = updated_stage.stage
        else:
            raise ZeusBSOValueError(
                f"Unknown job_id ({updated_stage.job_id}). Expecting {self.fetched_job.job_id}"
            )

    def update_min(self, updated_min: UpdateJobMinCost) -> None:
        """Update exploration min training cost and corresponding batch size on fetched job.

        Args:
            updated_min: Job Id, new min cost and batch size.
        """
        if self.fetched_job is None:
            raise ZeusBSOServiceBadOperationError("No job is fetched.")

        if self.fetched_job.job_id == updated_min.job_id:
            self.fetched_job.min_cost = updated_min.min_cost
            self.fetched_job.min_cost_batch_size = updated_min.min_cost_batch_size
        else:
            raise ZeusBSOValueError(
                f"Unknown job_id ({updated_min.job_id}). Expecting {self.fetched_job.job_id}"
            )

    def update_generator_state(self, updated_state: UpdateGeneratorState) -> None:
        """Update generator state on fetched job.

        Args:
            updated_state: Job Id and new generator state.
        """
        if self.fetched_job is None:
            raise ZeusBSOServiceBadOperationError("No job is fetched.")

        if self.fetched_job.job_id == updated_state.job_id:
            self.fetched_job.mab_random_generator_state = updated_state.state
        else:
            raise ZeusBSOValueError(
                f"Unknown job_id ({updated_state.job_id}). Expecting {self.fetched_job.job_id}"
            )

    def create_job(self, new_job: CreateJob) -> None:
        """Create a new job by adding a new job to the session.

        Args:
            new_job: Job configuration for a new job.
        """
        self.session.add(new_job.to_orm())

    def check_job_fetched(self, job_id: str) -> bool:
        """Check if this job is already fetched before.

        Args:
            job_id: Job id.

        Returns:
            True if this job was fetched and in session. Otherwise, return false.
        """
        return not (self.fetched_job is None or self.fetched_job.job_id != job_id)

    async def delete_job(self, job_id: str) -> bool:
        """Delete the job of a given job_Id.

        Args:
            job_id: Job id.

        Returns:
            True if the job got deleted.
        """
        stmt = select(JobTable).where(JobTable.job_id == job_id)
        job = await self.session.scalar(stmt)

        if job is None:
            return False

        # We can't straight delete using a query, since some db such as sqlite
        # Foreign Key is default to OFF, so "on delete = cascade" will not be fired.
        await self.session.delete(job)
        return True



"""Pipeline frequency optimizer server FastAPI router."""

from __future__ import annotations

import logging
from typing import Callable

from fastapi import Depends, FastAPI, Response, Request
from fastapi.routing import APIRoute

from zeus.utils.logging import get_logger
from zeus.optimizer.pipeline_frequency.common import (
    REGISTER_JOB_URL,
    REGISTER_RANK_URL,
    GET_FREQUENCY_SCHEDULE_URL,
    REPORT_PROFILING_RESULT_URL,
    JobInfo,
    RankInfo,
    PFOServerSettings,
    ProfilingResult,
    FrequencySchedule,
)
from zeus.optimizer.pipeline_frequency.server.job_manager import (
    JobManager,
    init_global_job_manager,
    get_global_job_manager,
)

logger = get_logger(__name__)
app = FastAPI()


class LoggingRoute(APIRoute):
    """Route handler that logs out all requests and responses in DEBUG level."""

    def get_route_handler(self) -> Callable:
        """Wrap the original handler with debug messages."""
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            response: Response = await original_route_handler(request)
            logger.debug(
                "%s %s: %s -> %s",
                request.method,
                request.url,
                await request.json() if await request.body() else "None",
                response.body.decode(response.charset),
            )
            return response

        return custom_route_handler


settings = PFOServerSettings()
logging.basicConfig(level=logging.getLevelName(settings.log_level))
if logging.getLevelName(settings.log_level) <= logging.DEBUG:
    app.router.route_class = LoggingRoute


@app.on_event("startup")
async def startup_hook():
    """Startup hook."""
    logger.info("Using scheduler `%s`", settings.scheduler.__name__)
    init_global_job_manager(settings)


@app.post(REGISTER_JOB_URL, response_model=str)
async def register_job(
    job_info: JobInfo, job_manager: JobManager = Depends(get_global_job_manager)
) -> str:
    """Register the training job's information in the server."""
    job_info.set_job_id(scheduler_name=settings.scheduler.__name__)
    job_manager.register_job(job_info)
    return job_info.job_id


@app.post(REGISTER_RANK_URL)
async def register_rank(
    job_id: str,
    rank_info: RankInfo,
    job_manager: JobManager = Depends(get_global_job_manager),
) -> None:
    """Register each rank's information in the server."""
    job_manager.register_rank(job_id, rank_info)


@app.get(GET_FREQUENCY_SCHEDULE_URL, response_model=FrequencySchedule)
async def get_frequency_schedule(
    job_id: str,
    rank: int,
    job_manager: JobManager = Depends(get_global_job_manager),
) -> FrequencySchedule:
    """Return the next frequency schedule for the rank."""
    return await job_manager.get_frequency_schedule(job_id, rank)


@app.post(REPORT_PROFILING_RESULT_URL)
async def report_profiling_result(
    job_id: str,
    profiling_result: ProfilingResult,
    job_manager: JobManager = Depends(get_global_job_manager),
) -> None:
    """Report the profiling result for the most recent frequency schedule."""
    job_manager.report_profiling_result(job_id, profiling_result)



"""Controller that sets the GPU's frequency in a non-blocking fashion."""

from __future__ import annotations

import atexit
import contextlib
import multiprocessing as mp

from zeus.device import get_gpus
from zeus.device.gpu import ZeusGPUNotSupportedError


class FrequencyController:
    """Spawns a separate process that sets the GPU frequency."""

    def __init__(self, device_id: int = 0) -> None:
        """Instantiate the frequency controller.

        Args:
            device_id: Device ID of the GPU to control.
        """
        self._q: mp.Queue[int | None] = mp.Queue()
        self._proc = mp.Process(target=self._controller_process, args=(device_id,))

        atexit.register(self.end)
        self._proc.start()

    def set_frequency(self, frequency: int) -> None:
        """Set the GPU's frequency asynchronously.

        If `frequency` is zero, returns without doing anything.
        """
        if frequency != 0:
            self._q.put(frequency, block=False)

    def end(self) -> None:
        """Stop the controller process."""
        self._q.put(None, block=False)

    def _controller_process(self, device_id: int) -> None:
        """Receive frequency values through a queue and apply it."""
        gpus = get_gpus()
        # Return the power limit to the default.
        gpus.resetPowerManagementLimit(device_id)

        # Set the memory frequency to be the highest.
        max_mem_freq = max(gpus.getSupportedMemoryClocks(device_id))
        with contextlib.suppress(ZeusGPUNotSupportedError):
            gpus.setMemoryLockedClocks(device_id, max_mem_freq, max_mem_freq)

        # Set the SM frequency to be the highest.
        max_freq = max(gpus.getSupportedGraphicsClocks(device_id, max_mem_freq))
        gpus.setGpuLockedClocks(device_id, max_freq, max_freq)
        current_freq = max_freq

        # Wait on the queue for the next frequency to set.
        while True:
            target_freq = self._q.get(block=True)
            if target_freq is None:
                break
            if current_freq != target_freq:
                gpus.setGpuLockedClocks(device_id, target_freq, target_freq)
                current_freq = target_freq

        # Reset everything.
        with contextlib.suppress(ZeusGPUNotSupportedError):
            gpus.resetMemoryLockedClocks(device_id)
        gpus.resetGpuLockedClocks(device_id)



"""AMD GPUs."""

from __future__ import annotations
import functools
import os
import contextlib
from typing import Sequence
from functools import lru_cache

try:
    import amdsmi  # type: ignore
except ImportError:

    class MockAMDSMI:
        """Mock class for AMD SMI library."""

        def __getattr__(self, name):
            """Raise an error if any method is called.

            Since this class is only used when `amdsmi` is not available,
            something has gone wrong if any method is called.
            """
            raise RuntimeError(
                f"amdsmi is not available and amdsmi.{name} shouldn't have been called. "
                "This is a bug."
            )

    amdsmi = MockAMDSMI()

import zeus.device.gpu.common as gpu_common
from zeus.utils.logging import get_logger

logger = get_logger(name=__name__)


@lru_cache(maxsize=1)
def amdsmi_is_available() -> bool:
    """Check if amdsmi is available."""
    try:
        import amdsmi  # type: ignore
    except ImportError:
        logger.info("amdsmi is not available.")
        return False
    try:
        amdsmi.amdsmi_init()
        logger.info("amdsmi is available and initialized")
        return True
    except amdsmi.AmdSmiLibraryException as e:
        logger.info("amdsmi is available but could not initialize: %s", e)
        return False


def _handle_amdsmi_errors(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except amdsmi.AmdSmiLibraryException as e:
            exception_class = AMDGPU._exception_map.get(
                e.get_error_code(), gpu_common.ZeusGPUUnknownError
            )
            raise exception_class(e.get_error_info()) from e

    return wrapper


class AMDGPU(gpu_common.GPU):
    """Implementation of `GPU` for AMD GPUs."""

    def __init__(self, gpu_index: int) -> None:
        """Initialize the GPU object."""
        super().__init__(gpu_index)
        self._get_handle()
        # XXX(Jae-Won): Right now, the energy API's unit is broken (either the
        # `power` field or the `counter_resolution` field). Before that, we're
        # disabling the energy API.
        self._supportsGetTotalEnergyConsumption = False

    _exception_map = {
        1: gpu_common.ZeusGPUInvalidArgError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_INVAL
        2: gpu_common.ZeusGPUNotSupportedError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_NOT_SUPPORTED
        8: gpu_common.ZeusGPUTimeoutError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_TIMEOUT
        10: gpu_common.ZeusGPUNoPermissionError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_NO_PERM
        15: gpu_common.ZeusGPUMemoryError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_OUT_OF_RESOURCES
        18: gpu_common.ZeusGPUInitError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_INIT_ERROR
        31: gpu_common.ZeusGPUNotFoundError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_NOT_FOUND
        32: gpu_common.ZeusGPUInitError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_NOT_INIT
        34: gpu_common.ZeusGPUDriverNotLoadedError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_DRIVER_NOT_LOADED
        41: gpu_common.ZeusGPUInsufficientSizeError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_INSUFFICIENT_SIZE
        45: gpu_common.ZeusGPUDriverNotLoadedError,  # amdsmi.amdsmi_wrapper.AMDSMI_NO_ENERGY_DRV
        46: gpu_common.ZeusGPUDriverNotLoadedError,  # amdsmi.amdsmi_wrapper.AMDSMI_NO_MSR_DRV
        47: gpu_common.ZeusGPUDriverNotLoadedError,  # amdsmi.amdsmi_wrapper.AMDSMI_NO_HSMP_DRV
        48: gpu_common.ZeusGPUNotSupportedError,  # amdsmi.amdsmi_wrapper.AMDSMI_NO_HSMP_SUP
        49: gpu_common.ZeusGPUNotSupportedError,  # amdsmi.amdsmi_wrapper.AMDSMI_NO_HSMP_MSG_SUP
        50: gpu_common.ZeusGPUTimeoutError,  # amdsmi.amdsmi_wrapper.AMDSMI_HSMP_TIMEOUT
        51: gpu_common.ZeusGPUDriverNotLoadedError,  # amdsmi.amdsmi_wrapper.AMDSMI_NO_DRV
        52: gpu_common.ZeusGPULibraryNotFoundError,  # amdsmi.amdsmi_wrapper.AMDSMI_FILE_NOT_FOUND
        53: gpu_common.ZeusGPUInvalidArgError,  # amdsmi.amdsmi_wrapper.AMDSMI_ARG_PTR_NULL
        4294967295: gpu_common.ZeusGPUUnknownError,  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_UNKNOWN_ERROR
    }

    @_handle_amdsmi_errors
    def _get_handle(self):
        handles = amdsmi.amdsmi_get_processor_handles()
        if len(handles) <= self.gpu_index:
            raise gpu_common.ZeusGPUNotFoundError(
                f"GPU with index {self.gpu_index} not found. Found {len(handles)} GPUs."
            )
        self.handle = amdsmi.amdsmi_get_processor_handles()[self.gpu_index]

    @_handle_amdsmi_errors
    def getName(self) -> str:
        """Return the name of the GPU model."""
        info = amdsmi.amdsmi_get_gpu_asic_info(self.handle)
        return info["market_name"]

    @property
    def supports_nonblocking_setters(self) -> bool:
        """Return True if the GPU object supports non-blocking configuration setters."""
        return False

    @_handle_amdsmi_errors
    def getPowerManagementLimitConstraints(self) -> tuple[int, int]:
        """Return the minimum and maximum power management limits. Units: mW."""
        info = amdsmi.amdsmi_get_power_cap_info(self.handle)  # Returns in W
        return (info["min_power_cap"] * 1000, info["max_power_cap"] * 1000)

    @_handle_amdsmi_errors
    def setPowerManagementLimit(self, power_limit_mw: int, _block: bool = True) -> None:
        """Set the GPU's power management limit. Unit: mW."""
        amdsmi.amdsmi_set_power_cap(
            self.handle, 0, int(power_limit_mw * 1000)
        )  # Units for set_power_cap: microwatts

    @_handle_amdsmi_errors
    def resetPowerManagementLimit(self, _block: bool = True) -> None:
        """Reset the GPU's power management limit to the default value."""
        info = amdsmi.amdsmi_get_power_cap_info(self.handle)  # Returns in W
        amdsmi.amdsmi_set_power_cap(
            self.handle, 0, cap=int(info["default_power_cap"] * 1e6)
        )  # expects value in microwatts

    @_handle_amdsmi_errors
    def setPersistenceMode(self, enabled: bool, _block: bool = True) -> None:
        """Set persistence mode."""
        raise gpu_common.ZeusGPUNotSupportedError(
            "Persistence mode is not supported on AMD GPUs."
        )

    @_handle_amdsmi_errors
    def getSupportedMemoryClocks(self) -> list[int]:
        """Return a list of supported memory clock frequencies. Units: MHz."""
        info = amdsmi.amdsmi_get_clock_info(
            self.handle, amdsmi.AmdSmiClkType.MEM
        )  # returns MHz
        return [info["max_clk"], info["min_clk"]]

    @_handle_amdsmi_errors
    def setMemoryLockedClocks(
        self, min_clock_mhz: int, max_clock_mhz: int, _block: bool = True
    ) -> None:
        """Lock the memory clock to a specified range. Units: MHz."""
        amdsmi.amdsmi_set_gpu_clk_range(
            self.handle,
            min_clock_mhz,
            max_clock_mhz,
            clk_type=amdsmi.AmdSmiClkType.MEM,
        )

    @_handle_amdsmi_errors
    def resetMemoryLockedClocks(self, _block: bool = True) -> None:
        """Reset the locked memory clocks to the default."""
        # Get default MEM clock values
        info = amdsmi.amdsmi_get_clock_info(
            self.handle, amdsmi.AmdSmiClkType.MEM
        )  # returns MHz

        amdsmi.amdsmi_set_gpu_clk_range(
            self.handle,
            info["min_clk"],
            info["max_clk"],
            clk_type=amdsmi.AmdSmiClkType.MEM,
        )  # expects MHz

    @_handle_amdsmi_errors
    def getSupportedGraphicsClocks(
        self, memory_clock_mhz: int | None = None
    ) -> list[int]:
        """Return a list of supported graphics clock frequencies. Units: MHz.

        Args:
            memory_clock_mhz: Memory clock frequency to use. Some GPUs have
                different supported graphics clocks depending on the memory clock.
        """
        pass
        info = amdsmi.amdsmi_get_clock_info(
            self.handle, amdsmi.AmdSmiClkType.GFX
        )  # returns MHz
        return [info["max_clk"], info["min_clk"]]

    @_handle_amdsmi_errors
    def setGpuLockedClocks(
        self, min_clock_mhz: int, max_clock_mhz: int, _block: bool = True
    ) -> None:
        """Lock the GPU clock to a specified range. Units: MHz."""
        amdsmi.amdsmi_set_gpu_clk_range(
            self.handle,
            min_clock_mhz,
            max_clock_mhz,
            clk_type=amdsmi.AmdSmiClkType.GFX,
        )

    @_handle_amdsmi_errors
    def resetGpuLockedClocks(self, _block: bool = True) -> None:
        """Reset the locked GPU clocks to the default."""
        # Get default GPU clock values
        info = amdsmi.amdsmi_get_clock_info(
            self.handle, amdsmi.AmdSmiClkType.GFX
        )  # returns MHz

        amdsmi.amdsmi_set_gpu_clk_range(
            self.handle,
            info["min_clk"],
            info["max_clk"],
            clk_type=amdsmi.AmdSmiClkType.GFX,
        )  # expects MHz

    @_handle_amdsmi_errors
    def getInstantPowerUsage(self) -> int:
        """Return the current power draw of the GPU. Units: mW."""
        # returns in W, convert to mW
        return int(
            amdsmi.amdsmi_get_power_info(self.handle)["average_socket_power"] * 1000
        )

    @_handle_amdsmi_errors
    def getAverageMemoryPowerUsage(self) -> int:
        """Return the average power usage of the GPU's memory. Units: mW."""
        raise gpu_common.ZeusGPUNotSupportedError(
            "Average memory power usage is not supported on AMD GPUs."
        )

    @_handle_amdsmi_errors
    def supportsGetTotalEnergyConsumption(self) -> bool:
        """Check if the GPU supports retrieving total energy consumption."""
        if self._supportsGetTotalEnergyConsumption is None:
            try:
                _ = amdsmi.amdsmi_get_energy_count(self.handle)
                self._supportsGetTotalEnergyConsumption = True
            except amdsmi.AmdSmiLibraryException as e:
                if (
                    e.get_error_code() == 2
                ):  # amdsmi.amdsmi_wrapper.AMDSMI_STATUS_NOT_SUPPORTED
                    self._supportsGetTotalEnergyConsumption = False
                else:
                    raise e

        return self._supportsGetTotalEnergyConsumption

    @_handle_amdsmi_errors
    def getTotalEnergyConsumption(self) -> int:
        """Return the total energy consumption of the GPU since driver load. Units: mJ."""
        info = amdsmi.amdsmi_get_energy_count(self.handle)
        return int(
            info["power"] / 1e3
        )  # returns in micro Joules, convert to mili Joules


class AMDGPUs(gpu_common.GPUs):
    """AMD GPU Manager object, containing individual AMDGPU objects, abstracting amdsmi calls and handling related exceptions.

    !!! Important
        Currently only ROCm >= 6.1 is supported.

    `HIP_VISIBLE_DEVICES` environment variable is respected if set.
    For example, if there are 4 GPUs on the node and `HIP_VISIBLE_DEVICES=0,2`,
    only GPUs 0 and 2 are instantiated. In this case, to access
    GPU of HIP index 0, use the index 0, and for HIP index 2, use the index 1.

    When `HIP_VISIBLE_DEVICES` is not set but `CUDA_VISIBLE_DEVICES` is set,
    `CUDA_VISIBLE_DEVICES` is honored as if it were `HIP_VISIBLE_DEVICES`.
    """

    def __init__(self, ensure_homogeneous: bool = False) -> None:
        """Initialize AMDSMI and sets up the GPUs.

        Args:
            ensure_homogeneous (bool): If True, ensures that all tracked GPUs have the same name.
        """
        try:
            amdsmi.amdsmi_init()
            self._init_gpus()
            if ensure_homogeneous:
                self._ensure_homogeneous()
        except amdsmi.AmdSmiException as e:
            exception_class = AMDGPU._exception_map.get(
                e.value, gpu_common.ZeusBaseGPUError
            )
            raise exception_class(e.msg) from e

    @property
    def gpus(self) -> Sequence[AMDGPU]:
        """Return a list of AMDGPU objects being tracked."""
        return self._gpus

    def _init_gpus(self) -> None:
        # Must respect `HIP_VISIBLE_DEVICES` (or `CUDA_VISIBLE_DEVICES`) if set
        if (visible_device := os.environ.get("HIP_VISIBLE_DEVICES")) is not None or (
            visible_device := os.environ.get("CUDA_VISIBLE_DEVICES")
        ) is not None:
            if not visible_device:
                raise gpu_common.ZeusGPUInitError(
                    "HIP_VISIBLE_DEVICES or CUDA_VISIBLE_DEVICES is set but empty. "
                    "You can use either one for AMD GPUs, but it should either be unset "
                    "or a comma-separated list of GPU indices."
                )
            visible_indices = [int(idx) for idx in visible_device.split(",")]
        else:
            visible_indices = list(range(len(amdsmi.amdsmi_get_processor_handles())))

        self._gpus = [AMDGPU(gpu_num) for gpu_num in visible_indices]

    def __del__(self) -> None:
        """Shut down AMDSMI."""
        with contextlib.suppress(amdsmi.AmdSmiException):
            amdsmi.amdsmi_shut_down()  # Ignore error on shutdown. Neccessary for proper cleanup and test functionality



"""Provides report/next_batch_size during pruning stage."""

from __future__ import annotations

from zeus.optimizer.batch_size.server.batch_size_state.commands import (
    CreateConcurrentTrial,
    CreateExplorationTrial,
    ReadTrial,
)
from zeus.optimizer.batch_size.server.batch_size_state.models import ExplorationsPerJob
from zeus.optimizer.batch_size.server.database.schema import TrialStatus
from zeus.optimizer.batch_size.server.exceptions import (
    ZeusBSOServerRuntimeError,
    ZeusBSOValueError,
)
from zeus.optimizer.batch_size.server.job.models import JobState
from zeus.optimizer.batch_size.server.services.service import ZeusService
from zeus.utils.logging import get_logger
from zeus.utils.metric import zeus_cost

logger = get_logger(__name__)


class PruningExploreManager:
    """Pruning manager that manges the batch size states in pruning stage."""

    def __init__(self, service: ZeusService):
        """Set up zeus service."""
        self.service = service

    async def next_batch_size(
        self,
        job: JobState,
        exploration_history: ExplorationsPerJob,
    ) -> ReadTrial | list[int]:
        """Find the next batch size to explore.

        Three cases possible.
        1. Pruninig Stage : There is a batch size that has not explored during the round.
        2. Concurrent job : There is an exploration with "Dispatched" state.
        3. Mab stage : All batch sizes have been explored and round is over.

        Args:
            job: state of the job
            exploration_history: all "succeeded" explorations that we have done for that job

        Returns:
            Return the batch size to use during Pruning stage.
            If Pruning stage was over, return None.

        Raises:
            `ZeusBSOValueError`: If the value is invalid. EX) default batch size is not in the converged batch size list.
        """
        batch_sizes = job.batch_sizes
        exp_default_bs = job.default_batch_size

        for round in range(job.num_pruning_rounds):
            converged_bs_list = []

            min_cost_of_round = float("inf")
            min_batch_size_of_round = 0

            batch_sizes.sort()
            idx = batch_sizes.index(exp_default_bs)
            down = sorted(batch_sizes[: idx + 1], reverse=True)
            up = sorted(batch_sizes[idx + 1 :])

            for bs_list in [down, up]:
                for bs in bs_list:
                    if (
                        bs in exploration_history.explorations_per_bs
                        and len(exploration_history.explorations_per_bs[bs]) > round
                    ):
                        # Already explored at this round
                        if (
                            exploration_history.explorations_per_bs[bs][round].status
                            == TrialStatus.Dispatched
                        ):
                            # We are waiting for the result of this exploration -> Concurrent job!
                            return await self.service.create_trial(
                                CreateConcurrentTrial(
                                    job_id=job.job_id,
                                    batch_size=job.min_cost_batch_size,
                                )
                            )

                        if not exploration_history.explorations_per_bs[bs][
                            round
                        ].converged:
                            # Failed to converge -> Go to next list or round
                            break
                        else:
                            # Training converged.
                            converged_bs_list.append(bs)

                            m = exploration_history.explorations_per_bs[bs][round]
                            if m.energy is None or m.time is None:
                                raise ZeusBSOValueError(
                                    "Energy or time is not available for the exploration."
                                )
                            cost = zeus_cost(
                                m.energy, m.time, job.eta_knob, job.max_power
                            )
                            if cost < min_cost_of_round:
                                min_cost_of_round = cost
                                min_batch_size_of_round = bs

                    else:
                        # Did not explore this round. Should explore!
                        return await self.service.create_trial(
                            CreateExplorationTrial(
                                job_id=job.job_id,
                                batch_size=bs,
                            )
                        )

            # We should go to next round. Update exp_default_bs and batch sizes!
            exp_default_bs = min_batch_size_of_round
            batch_sizes = converged_bs_list

            logger.info(
                "[PruningExploreManager] go to next round(%d) new default bs = %d converged bs list = %s",
                round,
                exp_default_bs,
                batch_sizes,
            )

            if len(batch_sizes) == 0:
                raise ZeusBSOServerRuntimeError(
                    "No converged batch sizes has observed. Reconfigure batch_sizes and re-launch the job."
                )
        # After going through pruning rounds, we couldn't find the bs. Should go to MAB stage, so return good batch_sizes.
        return sorted(batch_sizes)



"""Utilities for asyncio."""

from __future__ import annotations

import asyncio
import logging
import functools
from typing import Any, Coroutine, TypeVar

from zeus.utils.logging import get_logger

T = TypeVar("T")
default_logger = get_logger(__name__)


def create_task(
    coroutine: Coroutine[Any, Any, T],
    logger: logging.Logger | None = None,
) -> asyncio.Task[T]:
    """Create an `asyncio.Task` but ensure that exceptions are logged.

    Reference: https://quantlane.com/blog/ensure-asyncio-task-exceptions-get-logged/

    Args:
        coroutine: The coroutine to be wrapped.
        logger: The logger to be used for logging exceptions. If `None`, the
            the logger with the name `zeus.utils.async_utils` is used.
    """
    loop = asyncio.get_running_loop()
    task = loop.create_task(coroutine)
    task.add_done_callback(
        functools.partial(_handle_task_exception, logger=logger or default_logger)
    )
    return task


def _handle_task_exception(task: asyncio.Task, logger: logging.Logger) -> None:
    """Print out exception and tracebook when a task dies with an exception."""
    try:
        task.result()
    except asyncio.CancelledError:
        # Cancellation should not be logged as an error.
        pass
    except Exception:
        # `logger.exception` automatically handles exception and traceback info.
        logger.exception("Job task died with an exception!")



from __future__ import annotations

import builtins
import os
import pytest
import platform

from unittest.mock import patch

from zeus.device.cpu import get_current_cpu_index


class MockFile:
    def __init__(self, content):
        self.content = content

    def read(self, *args, **kwargs):
        return self.content

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


@pytest.fixture
def mock_linux_files():
    STAT_CONTENT = (
        "6511 (cat) R 6501 6511 6501 34816 6511 4194304 95 0 0 0 0 "
        "0 0 0 20 0 1 0 7155 6053888 255 18446744073709551615 94288847015936 "
        "94288847031350 140727519907328 0 0 0 0 0 0 0 0 0 17 24 0 0 0 0 0 "
        "94288847043296 94288847044712 94288866566144 140727519913901 "
        "140727519913921 140727519913921 140727519916011 0"
    )
    PHYSICAL_PACKAGE_CONTENT = "0"

    mocked_stat_file = MockFile(STAT_CONTENT)
    mocked_phys_package_file = MockFile(PHYSICAL_PACKAGE_CONTENT)

    real_open = builtins.open

    def mock_file_open(filepath, *args, **kwargs):
        if filepath == "/proc/515/stat":
            return mocked_stat_file
        elif filepath == "/sys/devices/system/cpu/cpu24/topology/physical_package_id":
            return mocked_phys_package_file
        else:
            return real_open(filepath, *args, **kwargs)

    patch_open = patch("builtins.open", side_effect=mock_file_open)

    patch_open.start()

    yield

    patch_open.stop()


def test_get_current_cpu_index(mock_linux_files):
    assert get_current_cpu_index(515) == 0


@pytest.mark.skipif(platform.system() != "Linux", reason="Linux specific test")
def test_get_current_cpu_index_linux():
    assert platform.system() == "Linux"
    assert os.path.exists("/proc/cpuinfo")
    socket_ids = set()
    with open("/proc/cpuinfo") as f:
        for line in f:
            if "physical id" in line:
                socket_id = line.strip().split(":")[1].strip()
                socket_ids.add(socket_id)

    cpu_index = get_current_cpu_index()
    assert cpu_index >= 0 and cpu_index < len(socket_ids)



"""Zeus server exceptions."""

from zeus.exception import ZeusBaseError


class ZeusBSOServerBaseError(ZeusBaseError):
    """Base error class for BSO server."""

    def __init__(self, msg: str):
        """Set status code."""
        super().__init__(msg)
        self.status_code = 500


class ZeusBSOJobConfigMismatchError(ZeusBSOServerBaseError):
    """When the job configuration doesn't align for the same job_id."""

    def __init__(self, msg: str):
        """Set status code."""
        super().__init__(msg)
        self.status_code = 409


class ZeusBSOValueError(ZeusBSOServerBaseError):
    """When the certain value is invalid."""

    def __init__(self, msg: str):
        """Set status code."""
        super().__init__(msg)
        self.status_code = 400


class ZeusBSOServerNotFoundError(ZeusBSOServerBaseError):
    """Resource we are looking for is not found."""

    def __init__(self, msg: str):
        """Set status code."""
        super().__init__(msg)
        self.status_code = 404


class ZeusBSOServiceBadOperationError(ZeusBSOServerBaseError):
    """When the operation doesn't meet requirements. ex) fetching measurements before fetching a job."""

    def __init__(self, msg: str):
        """Set status code."""
        super().__init__(msg)
        self.status_code = 400


class ZeusBSOServerRuntimeError(ZeusBSOServerBaseError):
    """Initialization or other errors during runtime."""

    def __init__(self, msg: str):
        """Set status code."""
        super().__init__(msg)
        self.status_code = 500



"""Common utilities for device management."""

from __future__ import annotations

import os
import ctypes
from functools import lru_cache

from zeus.utils.logging import get_logger

logger = get_logger(__name__)


@lru_cache(maxsize=1)
def has_sys_admin() -> bool:
    """Check if the current process has `SYS_ADMIN` capabilities."""
    # First try to read procfs.
    try:
        with open("/proc/self/status") as f:
            for line in f:
                if line.startswith("CapEff"):
                    bitmask = int(line.strip().split()[1], 16)
                    has = bool(bitmask & (1 << 21))
                    logger.info(
                        "Read security capabilities from /proc/self/status -- SYS_ADMIN: %s",
                        has,
                    )
                    return has
    except Exception:
        logger.info("Failed to read capabilities from /proc/self/status", exc_info=True)

    # If that fails, try to use the capget syscall.
    class CapHeader(ctypes.Structure):
        _fields_ = [("version", ctypes.c_uint32), ("pid", ctypes.c_int)]

    class CapData(ctypes.Structure):
        _fields_ = [
            ("effective", ctypes.c_uint32),
            ("permitted", ctypes.c_uint32),
            ("inheritable", ctypes.c_uint32),
        ]

    # Attempt to load libc and set up capget
    try:
        libc = ctypes.CDLL("libc.so.6")
        capget = libc.capget
        capget.argtypes = [ctypes.POINTER(CapHeader), ctypes.POINTER(CapData)]
        capget.restype = ctypes.c_int
    except Exception:
        logger.info("Failed to load libc.so.6", exc_info=True)
        return False

    # Initialize the header and data structures
    header = CapHeader(version=0x20080522, pid=0)  # Use the current process
    data = CapData()

    # Call capget and check for errors
    if capget(ctypes.byref(header), ctypes.byref(data)) != 0:
        errno = ctypes.get_errno()
        logger.info(
            "capget failed with error: %s (errno %s)", os.strerror(errno), errno
        )
        return False

    bitmask = data.effective
    has = bool(bitmask & (1 << 21))
    logger.info("Read security capabilities from capget -- SYS_ADMIN: %s", has)
    return has



"""Zeus batch size optimizer client that communicates with server."""

from __future__ import annotations
import atexit

import httpx
from zeus.callback import Callback
from zeus.monitor import ZeusMonitor
from zeus.optimizer.batch_size.common import (
    GET_NEXT_BATCH_SIZE_URL,
    REGISTER_JOB_URL,
    REPORT_END_URL,
    REPORT_RESULT_URL,
    CreatedJob,
    JobSpecFromClient,
    JobSpec,
    TrialId,
    ReportResponse,
    TrainingResult,
)
from zeus.optimizer.batch_size.exceptions import (
    ZeusBSOBadOperationError,
    ZeusBSOConfigError,
    ZeusBSOOperationOrderError,
    ZeusBSORuntimError,
    ZeusBSOTrainFailError,
)
from zeus.utils.logging import get_logger
from zeus.device import get_gpus

logger = get_logger(__name__)


class BatchSizeOptimizer(Callback):
    """Batch size optimizer client that talks to server.

    The following methods must be called in order inside the training script:

    - `get_batch_size`: At the beginning of the script.
    - `on_train_begin`: Before running any epochs.
    - `on_evaluate`: After each epoch when the validation metric is available.

    One batch size optimizer per one training session of the job.
    The set of GPUs to be used for training should be homogeneous, and will be inferred
    from the `ZeusMonitor` instance passed into the constructor.
    """

    def __init__(
        self, monitor: ZeusMonitor, server_url: str, job: JobSpec, rank: int = 0
    ) -> None:
        """Initialize the optimizer, and register the job to the server.

        If job is already registered, check if the job configuration is identical with previously registered config.

        Args:
            monitor: `ZeusMonitor` instance configured to measure the energy of all and only the GPUs used for training.
            server_url: url of batch size optimizer server
            job: job specification. Refer to `JobSpec` for job specifcatio parameters.
            rank: rank of gpu in the case of distributed training. We only let rank = 0 gpu to request for a batch size.
        """
        self.monitor = monitor
        self.server_url = server_url
        self.cur_epoch = 0  # 0-indexed
        self.running_time = 0.0
        self.consumed_energy = 0.0
        self.training_finished = False
        self.trial_number = 0
        self.rank = rank

        # Currently, the BSO only supports homogeneous GPU training.
        gpus = get_gpus(ensure_homogeneous=True)
        if len(gpus) == 0:
            raise ZeusBSOConfigError("No GPUs detected.")

        # set gpu configurations(max_power, number of gpus, and gpu model)
        self.job_spec = JobSpecFromClient(
            **job.dict(),
            max_power=gpus.getPowerManagementLimitConstraints(0)[1]
            // 1000
            * len(monitor.gpu_indices),
            number_of_gpus=len(monitor.gpu_indices),
            gpu_model=gpus.getName(0),
        )

        # Track the batch size of current job
        self.current_batch_size = 0

        # Register job
        res = httpx.post(
            self.server_url + REGISTER_JOB_URL, content=self.job_spec.json()
        )
        self._handle_response(res)

        self.job = CreatedJob.parse_obj(res.json())

        logger.critical(
            "Job is registered with job_id: \x1b[31;1m%s\x1b[0m", self.job.job_id
        )
        logger.info("Job is registered: %s", str(self.job))

    def get_batch_size(self) -> int:
        """Get batch size to use from the BSO server.

        Returns:
            return a batch size to use for current job

        Raises:
            `ZeusBSORuntimError`: if the batch size we receive is invalid
        """
        if self.rank != 0:
            raise ZeusBSOBadOperationError("Only rank 0 gpu can ask for a batch size.")

        if self.current_batch_size != 0:
            # If we already got the batch size, return
            return self.current_batch_size

        self.cur_epoch = 0
        res = httpx.get(
            self.server_url + GET_NEXT_BATCH_SIZE_URL,
            params={"job_id": self.job.job_id},
        )
        self._handle_response(res)
        trial_id = TrialId.parse_obj(res.json())

        if trial_id.batch_size not in self.job.batch_sizes:
            raise ZeusBSORuntimError(
                f"Zeus server returned a strange batch_size: {trial_id.batch_size}"
            )

        self.current_batch_size = trial_id.batch_size
        self.trial_number = trial_id.trial_number

        logger.info("[BatchSizeOptimizer] Chosen batch size: %s", trial_id.batch_size)

        def report_end() -> None:
            httpx.patch(self.server_url + REPORT_END_URL, content=trial_id.json())

        atexit.register(report_end)
        return trial_id.batch_size

    def on_train_begin(self) -> None:
        """Start the monitor window and mark training is started."""
        self.training_finished = False
        self.monitor.begin_window("BatciSizeOptimizerClient")

    def on_evaluate(
        self,
        metric: float,
    ) -> None:
        """Determine whether or not to stop training after evaluation.

        Training stops when
        - `max_epochs` was reached, or
        - the target metric was reached. or
        - Cost exceeded the early stop threshold

        Args:
            metric: Validation metric of this epoch. See also `higher_metric_is_better` in [`JobParams`][zeus.optimizer.batch_size.common.JobParams].

        Raises:
            `ZeusBSOOperationOrderError`: When `get_batch_size` was not called first.
            `ZeusBSOTrainFailError`: When train failed for a chosen batch size and should be stopped.
                                    This batch size will not be tried again. To proceed training, re-launch the training then bso will select another batch size
            `ZeusBSORuntimError`: When the server returns an error
        """
        if self.current_batch_size == 0:
            raise ZeusBSOOperationOrderError(
                "Call get_batch_size to set the batch size first"
            )

        if self.training_finished:
            return

        self.cur_epoch += 1
        measurement = self.monitor.end_window("BatciSizeOptimizerClient")

        # Accumulate time and energy
        self.running_time += measurement.time
        self.consumed_energy += measurement.total_energy

        training_result = TrainingResult(
            job_id=self.job.job_id,
            batch_size=self.current_batch_size,
            trial_number=self.trial_number,
            time=self.running_time,
            energy=self.consumed_energy,
            metric=metric,
            current_epoch=self.cur_epoch,
        )

        # report to the server about the result of this training
        res = httpx.post(
            self.server_url + REPORT_RESULT_URL, content=training_result.json()
        )
        self._handle_response(res)

        parsed_response = ReportResponse.parse_obj(res.json())

        if not parsed_response.stop_train:
            # Should keep training. Re-open the window
            self.monitor.begin_window("BatciSizeOptimizerClient")
        else:
            # Train is over. If not converged, raise an error
            self.training_finished = True
            if not parsed_response.converged:
                raise ZeusBSOTrainFailError(
                    f"Train failed: {parsed_response.message} This batch size will not be selected again. Please re-launch the training"
                )

    def _handle_response(self, res: httpx.Response) -> None:
        """Check if the response is success. Otherwise raise an error with error message from the server.

        Args:
            res: response from the server
        """
        if not (200 <= (code := res.status_code) < 300):
            raise ZeusBSORuntimError(
                f"Zeus server returned status code {code}: {res.text}"
            )



"""Utilities for framework-specific code."""

from __future__ import annotations

import types
from typing import Literal
from functools import lru_cache

from zeus.utils.logging import get_logger

logger = get_logger(name=__name__)
MODULE_CACHE: dict[str, types.ModuleType] = {}


@lru_cache(maxsize=1)
def torch_is_available(ensure_available: bool = False, ensure_cuda: bool = True):
    """Check if PyTorch is available."""
    try:
        import torch

        cuda_available = torch.cuda.is_available()
        if ensure_cuda and not cuda_available:
            raise RuntimeError("PyTorch is available but does not have CUDA support.")
        MODULE_CACHE["torch"] = torch
        logger.info(
            "PyTorch %s CUDA support is available.",
            "with" if cuda_available else "without",
        )
        return True
    except ImportError as e:
        logger.info("PyTorch is not available.")
        if ensure_available:
            raise RuntimeError("Failed to import Pytorch") from e
        return False


@lru_cache(maxsize=1)
def jax_is_available(ensure_available: bool = False, ensure_cuda: bool = True):
    """Check if JAX is available."""
    try:
        import jax  # type: ignore

        cuda_available = jax.devices("gpu")
        if ensure_cuda and not cuda_available:
            raise RuntimeError("JAX is available but does not have CUDA support.")
        MODULE_CACHE["jax"] = jax
        logger.info(
            "JAX %s CUDA support is available.", "with" if cuda_available else "without"
        )
        return True
    except ImportError as e:
        logger.info("JAX is not available")
        if ensure_available:
            raise RuntimeError("Failed to import JAX") from e
        return False


def sync_execution(
    gpu_devices: list[int], sync_with: Literal["torch", "jax"] = "torch"
) -> None:
    """Block until all computations on the specified devices are finished.

    PyTorch only runs GPU computations asynchronously, so synchronizing computations
    for the given GPU devices is done by calling `torch.cuda.synchronize` on each
    device. On the other hand, JAX runs both CPU and GPU computations asynchronously,
    but by default it only has a single CPU device (id=0). Therefore for JAX, all GPU
    devices passed in and the CPU device (id=0) are synchronized.

    !!! Note
        `jax.device_put` with `block_until_ready` is used to synchronize computations
        on JAX devices. This is a workaround to the lack of a direct API for
        synchronizing computations on JAX devices. Tracking issue:
        https://github.com/google/jax/issues/4335

    !!! Note
        Across the Zeus library, an integer device index corresponds to a single whole
        physical device. This is usually what you want, except when using more advanced
        device partitioning (e.g., using `--xla_force_host_platform_device_count` in JAX
        to partition CPUs into more pieces). In such cases, you probably want to opt out
        from using this function and handle synchronization manually at the appropriate
        granularity.

    Args:
        gpu_devices: GPU device indices to synchronize.
        sync_with: Deep learning framework to use to synchronize computations.
            Defaults to `"torch"`, in which case `torch.cuda.synchronize` will be used.
    """
    if sync_with == "torch" and torch_is_available(ensure_available=True):
        torch = MODULE_CACHE["torch"]
        for device in gpu_devices:
            torch.cuda.synchronize(device)
        return

    if sync_with == "jax" and jax_is_available(ensure_available=True):
        jax = MODULE_CACHE["jax"]
        futures = [
            jax.device_put(0.0, device=jax.devices("gpu")[device]) + 0
            for device in gpu_devices
        ]
        futures.append(jax.device_put(0.0, device=jax.devices("cpu")[0]) + 0)
        jax.block_until_ready(futures)
        return

    raise RuntimeError("No framework is available.")



"""RAPL CPUs.

- RAPL (Running Average Power Limit):
   RAPL is a technology introduced by Intel that allows for power consumption monitoring and control at the processor and memory subsystem level. It provides mechanisms to enforce power limits and manage thermal conditions effectively.

- Power Zone:
   A power zone in the context of RAPL refers to a logical grouping of components within the CPU or system that share a common power domain. Each power zone can be monitored and controlled independently. Typical power zones include the entire package, specific cores, and memory subsystems.

- Package:
   The package refers to the physical CPU chip, which may contain multiple cores and integrated components. In RAPL, the package power domain encompasses the power consumption of all the cores and integrated units within the CPU package.

"""

from __future__ import annotations

import atexit
import multiprocessing as mp
import os
import time
import warnings
from functools import lru_cache
from glob import glob
from multiprocessing.sharedctypes import Synchronized
from typing import Sequence

import zeus.device.cpu.common as cpu_common
from zeus.device.cpu.common import CpuDramMeasurement
from zeus.device.exception import ZeusBaseCPUError
from zeus.utils.logging import get_logger

logger = get_logger(name=__name__)

RAPL_DIR = "/sys/class/powercap/intel-rapl"

# Location of RAPL files when in a docker container. See
# https://ml.energy/zeus/getting_started/#system-privileges for more details
CONTAINER_RAPL_DIR = "/zeus_sys/class/powercap/intel-rapl"


# Assuming a maximum power draw of 1000 Watts when we are polling every 0.1 seconds, the maximum
# amount the RAPL counter would increase
RAPL_COUNTER_MAX_INCREASE = 1000 * 1e6 * 0.1


class RaplWraparoundTracker:
    """Monitor the wrapping around of RAPL counters.

    This class acts as a lower level wrapper around a Python process that polls
    the wrapping of RAPL counters. This is primarily used by
    [`RAPLCPUs`][zeus.device.cpu.rapl.RAPLCPUs].

    !!! Warning
        Since the monitor spawns a child process, **it should not be instantiated as a global variable**.
        Python puts a protection to prevent creating a process in global scope.
        Refer to the "Safe importing of main module" section in the
        [Python documentation](https://docs.python.org/3/library/multiprocessing.html#the-spawn-and-forkserver-start-methods)
        for more details.

    Attributes:
        rapl_file_path (str): File path of rapl file to track wraparounds for.
        max_energy_uj (float): Max value of rapl counter for `rapl_file_path` file. Used to
            determine the sleep period between polls
    """

    def __init__(
        self,
        rapl_file_path: str,
        max_energy_uj: float,
    ) -> None:
        """Initialize the rapl monitor.

        Args:
            rapl_file_path: File path where the RAPL file is located
            max_energy_uj: Max energy range uj value
        """
        if not os.path.exists(rapl_file_path):
            raise ValueError(f"{rapl_file_path} is not a valid file path")

        # Set up logging.
        self.logger = get_logger(type(self).__name__)

        self.logger.info("Monitoring wrap around of %s", rapl_file_path)

        context = mp.get_context("spawn")
        self.wraparound_counter = context.Value("i", 0)
        # Spawn the power polling process.
        atexit.register(self._stop)
        self.process = context.Process(
            target=_polling_process,
            args=(rapl_file_path, max_energy_uj, self.wraparound_counter),
        )
        self.process.start()

    def _stop(self) -> None:
        """Stop monitoring power usage."""
        if self.process is not None:
            self.process.terminate()
            self.process.join(timeout=1.0)
            self.process.kill()
            self.process = None

    def get_num_wraparounds(self) -> int:
        """Get the number of wraparounds detected by the polling process."""
        with self.wraparound_counter.get_lock():
            return self.wraparound_counter.value


def _polling_process(
    rapl_file_path: str, max_energy_uj: float, wraparound_counter: Synchronized[int]
) -> None:
    """Check for wraparounds in the specified rapl file."""
    try:
        with open(rapl_file_path) as rapl_file:
            last_energy_uj = float(rapl_file.read().strip())
        while True:
            sleep_time = 1.0
            with open(rapl_file_path, "r") as rapl_file:
                energy_uj = float(rapl_file.read().strip())
            if max_energy_uj - energy_uj < RAPL_COUNTER_MAX_INCREASE:
                sleep_time = 0.1
            if energy_uj < last_energy_uj:
                with wraparound_counter.get_lock():
                    wraparound_counter.value += 1
            last_energy_uj = energy_uj
            time.sleep(sleep_time)
    except KeyboardInterrupt:
        return


@lru_cache(maxsize=1)
def rapl_is_available() -> bool:
    """Check if RAPL is available."""
    if not os.path.exists(RAPL_DIR) and not os.path.exists(CONTAINER_RAPL_DIR):
        logger.info("RAPL is not supported on this CPU.")
        return False
    logger.info("RAPL is available.")
    return True


class ZeusRAPLNotSupportedError(ZeusBaseCPUError):
    """Zeus CPU exception class wrapper for RAPL not supported on CPU."""

    def __init__(self, message: str) -> None:
        """Initialize Zeus Exception."""
        super().__init__(message)


class ZeusRAPLFileInitError(ZeusBaseCPUError):
    """Zeus CPU exception class wrapper for RAPL file initialization error on CPU."""

    def __init__(self, message: str) -> None:
        """Initialize Zeus Exception."""
        super().__init__(message)


class ZeusRAPLPermissionError(ZeusBaseCPUError):
    """Zeus GPU exception that wraps No Permission to perform GPU operation."""

    def __init__(self, message: str) -> None:
        """Intialize the exception object."""
        super().__init__(message)


class RAPLFile:
    """RAPL File class for each RAPL file.

    This class defines the interface for interacting with a RAPL file for a package. A package can
    be a CPU or DRAM
    """

    def __init__(self, path: str) -> None:
        """Initialize RAPL file object. Each RAPL file object manages one energy_uj file."""
        self.path = path
        self.energy_uj_path = os.path.join(path, "energy_uj")
        try:
            with open(os.path.join(path, "name"), "r") as name_file:
                self.name: str = name_file.read().strip()
        except FileNotFoundError as err:
            raise ZeusRAPLFileInitError("Error reading package name") from err
        try:
            with open(self.energy_uj_path) as energy_file:
                self.last_energy = float(energy_file.read().strip())
        except FileNotFoundError as err:
            raise ZeusRAPLFileInitError("Error reading package energy") from err
        except PermissionError as err:
            raise cpu_common.ZeusCPUNoPermissionError(
                "Can't read file due to permission error"
            ) from err
        try:
            with open(
                os.path.join(path, "max_energy_range_uj"), "r"
            ) as max_energy_file:
                self.max_energy_range_uj = float(max_energy_file.read().strip())
        except FileNotFoundError as err:
            raise ZeusRAPLFileInitError(
                "Error reading package max energy range"
            ) from err

        self.wraparound_tracker = RaplWraparoundTracker(
            self.energy_uj_path, self.max_energy_range_uj
        )

    def __str__(self) -> str:
        """Return a string representation of the RAPL file object."""
        return f"RAPLFile(Path: {self.path}\nEnergy_uj_path: {self.energy_uj_path}\nName: {self.name}\
        \nLast_energy: {self.last_energy}\nMax_energy: {self.max_energy_range_uj})"

    def read(self) -> float:
        """Read the current energy value from the energy_uj file.

        Returns:
            The current energy value in millijoules.
        """
        with open(self.energy_uj_path) as energy_file:
            new_energy_uj = float(energy_file.read().strip())
        num_wraparounds = self.wraparound_tracker.get_num_wraparounds()
        return (new_energy_uj + num_wraparounds * self.max_energy_range_uj) / 1000.0


class RAPLCPU(cpu_common.CPU):
    """Control a single CPU that supports RAPL."""

    def __init__(self, cpu_index: int, rapl_dir: str) -> None:
        """Initialize the Intel CPU with a specified index."""
        super().__init__(cpu_index)
        self.rapl_dir = rapl_dir
        self._get_powerzone()

    _exception_map = {
        FileNotFoundError: cpu_common.ZeusCPUNotFoundError,
        PermissionError: cpu_common.ZeusCPUNoPermissionError,
        OSError: cpu_common.ZeusCPUInitError,
    }

    def _get_powerzone(self) -> None:
        self.path = os.path.join(self.rapl_dir, f"intel-rapl:{self.cpu_index}")
        self.rapl_file: RAPLFile = RAPLFile(self.path)
        self.dram: RAPLFile | None = None
        for dir in os.listdir(self.path):
            if "intel-rapl" in dir:
                try:
                    rapl_file = RAPLFile(os.path.join(self.path, dir))
                except ZeusRAPLFileInitError as err:
                    warnings.warn(
                        f"Failed to initialize subpackage {err}", stacklevel=1
                    )
                    continue
                if rapl_file.name == "dram":
                    self.dram = rapl_file

    def getTotalEnergyConsumption(self) -> CpuDramMeasurement:
        """Returns the total energy consumption of the specified powerzone. Units: mJ."""
        cpu_mj = self.rapl_file.read()
        dram_mj = None
        if self.dram is not None:
            dram_mj = self.dram.read()
        return CpuDramMeasurement(cpu_mj=cpu_mj, dram_mj=dram_mj)

    def supportsGetDramEnergyConsumption(self) -> bool:
        """Returns True if the specified CPU powerzone supports retrieving the subpackage energy consumption."""
        return self.dram is not None


class RAPLCPUs(cpu_common.CPUs):
    """RAPL CPU Manager object, containing individual RAPLCPU objects, abstracting RAPL calls and handling related exceptions."""

    def __init__(self) -> None:
        """Instantiates IntelCPUs object, setting up tracking for specified Intel CPUs."""
        if not rapl_is_available():
            raise ZeusRAPLNotSupportedError("RAPL is not supported on this CPU.")

        self.rapl_dir = RAPL_DIR if os.path.exists(RAPL_DIR) else CONTAINER_RAPL_DIR
        self._init_cpus()

    @property
    def cpus(self) -> Sequence[RAPLCPU]:
        """Returns a list of CPU objects being tracked."""
        return self._cpus

    def _init_cpus(self) -> None:
        """Initialize all Intel CPUs."""
        self._cpus = []

        def sort_key(dir):
            return int(dir.split(":")[1])

        for dir in sorted(glob(f"{self.rapl_dir}/intel-rapl:*"), key=sort_key):
            parts = dir.split(":")
            if len(parts) > 1 and parts[1].isdigit():
                self._cpus.append(RAPLCPU(int(parts[1]), self.rapl_dir))

    def __del__(self) -> None:
        """Shuts down the Intel CPU monitoring."""
        pass



"""Generate source code references pages."""

from pathlib import Path

import mkdocs_gen_files

REF_DIR = Path("reference")


nav = mkdocs_gen_files.Nav()

for path in sorted(Path("zeus").rglob("*.py")):
    # Path to the generated markdown file.
    doc_path = path.relative_to("zeus").with_suffix(".md")

    # Skip BSO server migration-related files.
    if str(doc_path).find("batch_size/migrations") != -1:
        continue

    full_doc_path = REF_DIR / doc_path
    module_path = path.with_suffix("")
    parts = tuple(module_path.parts)

    # zeus/run/__init__ can just be zeus/run.
    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    # We currently don't have __main__ modules, but just in case.
    elif parts[-1] == "__main__":
        continue

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        identifier = ".".join(parts)
        fd.write(f"::: {identifier}\n")

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_list = list(nav.build_literate_nav())
    nav_list[0] = "* [Source Code Reference](index.md)\n"
    nav_file.writelines(nav_list)



"""Managing database connection.

Heavily inspired by https://praciano.com.br/fastapi-and-async-sqlalchemy-20-with-pytest-done-right.html
and https://medium.com/@tclaitken/setting-up-a-fastapi-app-with-async-sqlalchemy-2-0-pydantic-v2-e6c540be4308
"""

from __future__ import annotations

import contextlib
from typing import Any, AsyncIterator

from sqlalchemy.ext.asyncio import (
    AsyncConnection,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from zeus.optimizer.batch_size.server.config import settings
from zeus.optimizer.batch_size.server.exceptions import ZeusBSOServerRuntimeError


class DatabaseSessionManager:
    """Session manager class."""

    def __init__(self, host: str, engine_kwargs: dict[str, Any] | None = None):
        """Create async engine and session maker."""
        if engine_kwargs is None:
            engine_kwargs = {}
        self._engine = create_async_engine(host, **engine_kwargs)
        self._sessionmaker = async_sessionmaker(autocommit=False, bind=self._engine)

    async def close(self):
        """Close connection."""
        if self._engine is None:
            raise ZeusBSOServerRuntimeError("DatabaseSessionManager is not initialized")
        await self._engine.dispose()

        self._engine = None
        self._sessionmaker = None

    @contextlib.asynccontextmanager
    async def connect(self) -> AsyncIterator[AsyncConnection]:
        """Connect to db."""
        if self._engine is None:
            raise ZeusBSOServerRuntimeError("DatabaseSessionManager is not initialized")

        async with self._engine.begin() as connection:
            try:
                yield connection
            except Exception:
                await connection.rollback()
                raise

    @contextlib.asynccontextmanager
    async def session(self) -> AsyncIterator[AsyncSession]:
        """Get session from session maker."""
        if self._sessionmaker is None:
            raise ZeusBSOServerRuntimeError("DatabaseSessionManager is not initialized")

        session = self._sessionmaker()
        try:
            yield session
        except Exception:
            await session.rollback()
            raise
        finally:
            await session.close()


# Initialize session manager.
sessionmanager = DatabaseSessionManager(
    settings.database_url, {"echo": settings.echo_sql}
)


async def get_db_session() -> AsyncIterator[AsyncSession]:
    """Get db session from session manager. Used with fastapi dependency injection."""
    async with sessionmanager.session() as session:
        yield session



"""Infrastructure for calling callbacks."""

from __future__ import annotations


class Callback:
    """Base class for callbacks."""

    def on_train_begin(self) -> None:
        """Called at the beginning of training."""

    def on_train_end(self) -> None:
        """Called at the end of training."""

    def on_epoch_begin(self) -> None:
        """Called at the beginning of each epoch."""

    def on_epoch_end(self) -> None:
        """Called at the end of each epoch."""

    def on_step_begin(self) -> None:
        """Called at the beginning of each step."""

    def on_step_end(self) -> None:
        """Called at the end of each step."""

    def on_evaluate(self, metric: float) -> None:
        """Called after evaluating the model."""

    def on_instruction_begin(self, name: str) -> None:
        """Called at the beginning of pipeline instructions like forward or backward."""

    def on_instruction_end(self, name: str) -> None:
        """Called at the end of pipeline instructions like forward or backward."""


class CallbackSet(Callback):
    """A set of callbacks."""

    def __init__(self, callbacks: list[Callback]) -> None:
        """Initialize the callback set."""
        self.callbacks = callbacks

    def on_train_begin(self) -> None:
        """Called at the beginning of training."""
        for callback in self.callbacks:
            callback.on_train_begin()

    def on_train_end(self) -> None:
        """Called at the end of training."""
        for callback in self.callbacks:
            callback.on_train_end()

    def on_epoch_begin(self) -> None:
        """Called at the beginning of each epoch."""
        for callback in self.callbacks:
            callback.on_epoch_begin()

    def on_epoch_end(self) -> None:
        """Called at the end of each epoch."""
        for callback in self.callbacks:
            callback.on_epoch_end()

    def on_step_begin(self) -> None:
        """Called at the beginning of each step."""
        for callback in self.callbacks:
            callback.on_step_begin()

    def on_step_end(self) -> None:
        """Called at the end of each step."""
        for callback in self.callbacks:
            callback.on_step_end()

    def on_evaluate(self, metric: float) -> None:
        """Called after evaluating the model."""
        for callback in self.callbacks:
            callback.on_evaluate(metric)

    def on_instruction_begin(self, name: str) -> None:
        """Called at the beginning of pipeline instructions like forward or backward."""
        for callback in self.callbacks:
            callback.on_instruction_begin(name)

    def on_instruction_end(self, name: str) -> None:
        """Called at the end of pipeline instructions like forward or backward."""
        for callback in self.callbacks:
            callback.on_instruction_end(name)



from __future__ import annotations

import logging
import re
import uuid
from math import isclose

import pytest
from zeus.optimizer.batch_size.common import (
    GET_NEXT_BATCH_SIZE_URL,
    REGISTER_JOB_URL,
    REPORT_RESULT_URL,
    TrainingResult,
    TrialId,
)
from zeus.utils.metric import zeus_cost


@pytest.fixture(scope="session", autouse=True)
def logger_setup():
    logger = logging.getLogger(
        "zeus.optimizer.batch_size.server.mab"
    )  # for testing, propagate the log to the root logger so that caplog can capture
    logger.propagate = True
    yield


@pytest.mark.usefixtures("client")
@pytest.mark.usefixtures("caplog")
class TestPruningExploreManager:
    """Unit test class for pruning exploration."""

    batch_sizes: list[int] = [8, 16, 32, 64, 128, 256]

    def exploration_to_training_result(
        self,
        exploration: tuple[int, float, bool],
        job_id: str,
        trial_number: int,
        max_power: int,
    ) -> TrainingResult:
        energy = 1
        res = TrainingResult(
            job_id=job_id,
            batch_size=exploration[0],
            trial_number=trial_number,
            error=False,
            time=(2 * exploration[1] - energy) / max_power,
            energy=energy,
            metric=0.55 if exploration[2] else 0.4,
            current_epoch=100,
        )
        assert isclose(
            zeus_cost(res.energy, res.time, 0.5, max_power),
            exploration[1],
        )
        return res

    def register_job_with_default_bs(
        self, client, default_bs: int, helpers
    ) -> tuple[str, int]:
        job_id = f"test-{str(uuid.uuid4())}"
        fake_job = helpers.get_fake_job_config(job_id)
        fake_job["beta_knob"] = None
        fake_job["job_id"] = job_id
        fake_job["batch_sizes"] = self.batch_sizes
        fake_job["default_batch_size"] = default_bs

        response = client.post(REGISTER_JOB_URL, json=fake_job)
        assert response.status_code == 201

        return job_id, fake_job["max_power"]

    def run_exploration(
        self,
        client,
        caplog,
        job_id: str,
        exploration: list[tuple[int, float, bool]],
        result: list[int],
        max_power: int,
    ) -> None:
        """Drive the pruning explore manager and check results."""
        caplog.set_level(logging.INFO)

        for exp in exploration:
            response = client.get(
                GET_NEXT_BATCH_SIZE_URL,
                params={"job_id": job_id},
            )
            assert response.status_code == 200
            parsed_res = TrialId.parse_obj(response.json())
            assert (
                parsed_res.batch_size == exp[0]
            ), f"Expected {exp[0]} but got {parsed_res.batch_size} ({exp})"

            training_result = self.exploration_to_training_result(
                exp, job_id, parsed_res.trial_number, max_power
            )
            response = client.post(
                REPORT_RESULT_URL,
                content=training_result.json(),
            )
            assert response.status_code == 200, response.text
            assert response.json()["converged"] == exp[2]
            print(response.json()["message"])
        # Now good_bs should be equal to result!

        # this will construct mab
        response = client.get(
            GET_NEXT_BATCH_SIZE_URL,
            params={"job_id": job_id},
        )
        assert response.status_code == 200

        # Capture list of arms from stdout
        matches = re.search(r"with arms \[(.*?)\]", caplog.text)

        if matches:
            arms = [int(x) for x in matches.group(1).split(",")]
            arms.sort()
            assert arms == result
        else:
            assert False, "No output found from constructing Mab"

    def test_normal(self, client, caplog, helpers):
        """Test a typical case."""
        job_id, max_power = self.register_job_with_default_bs(client, 128, helpers)

        exploration = [
            (128, 10.0, True),
            (64, 9.0, True),
            (32, 8.0, True),
            (16, 12.0, True),
            (8, 21.0, False),
            (256, 15.0, True),
            (32, 8.0, True),
            (16, 12.0, False),
            (64, 9.0, True),
            (128, 10.0, True),
            (256, 17.0, False),
        ]

        result = [32, 64, 128]
        self.run_exploration(client, caplog, job_id, exploration, result, max_power)

    def test_default_is_largest(self, client, caplog, helpers):
        """Test the case when the default batch size is the largest one."""
        job_id, max_power = self.register_job_with_default_bs(client, 256, helpers)

        exploration = [
            (256, 7.0, True),
            (128, 8.0, True),
            (64, 9.0, True),
            (32, 13.0, True),
            (16, 22.0, False),
            (256, 8.0, True),
            (128, 8.5, True),
            (64, 9.0, True),
            (32, 12.0, True),
        ]
        result = [32, 64, 128, 256]
        self.run_exploration(client, caplog, job_id, exploration, result, max_power)

    def test_default_is_smallest(self, client, caplog, helpers):
        """Test the case when the default batch size is the smallest one."""
        job_id, max_power = self.register_job_with_default_bs(client, 8, helpers)

        exploration = [
            (8, 10.0, True),
            (16, 17.0, True),
            (32, 20.0, True),
            (64, 25.0, False),
            (8, 10.0, True),
            (16, 21.0, False),
        ]
        result = [8]
        self.run_exploration(client, caplog, job_id, exploration, result, max_power)

    def test_all_converge(self, client, caplog, helpers):
        """Test the case when every batch size converges."""
        job_id, max_power = self.register_job_with_default_bs(client, 64, helpers)
        exploration = [
            (64, 10.0, True),
            (32, 8.0, True),
            (16, 12.0, True),
            (8, 15.0, True),
            (128, 12.0, True),
            (256, 13.0, True),
            (32, 7.0, True),
            (16, 10.0, True),
            (8, 15.0, True),
            (64, 10.0, True),
            (128, 12.0, True),
            (256, 13.0, True),
        ]
        result = self.batch_sizes
        self.run_exploration(client, caplog, job_id, exploration, result, max_power)

    def test_every_bs_is_bs(self, client, caplog, helpers):
        """Test the case when every batch size other than the default fail to converge."""
        job_id, max_power = self.register_job_with_default_bs(client, 64, helpers)
        exploration = [
            (64, 10.0, True),
            (32, 22.0, False),
            (128, 25.0, False),
            (64, 9.0, True),
        ]
        result = [64]
        self.run_exploration(client, caplog, job_id, exploration, result, max_power)



"""Database schema."""

from __future__ import annotations

import enum
from datetime import datetime
from typing import Optional

from sqlalchemy import (
    Boolean,
    DateTime,
    Enum,
    Float,
    ForeignKey,
    ForeignKeyConstraint,
    Integer,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship
from sqlalchemy.sql.sqltypes import VARCHAR
from zeus.optimizer.batch_size.server.job.models import Stage


class Base(DeclarativeBase):
    """Base class for schemas."""

    pass


class JobTable(Base):
    """Job table schema.

    Refer to [`JobState`][zeus.optimizer.batch_size.server.job.models.JobState] for attributes.
    """

    __tablename__ = "Job"

    job_id: Mapped[str] = mapped_column(VARCHAR(400), primary_key=True)
    job_id_prefix: Mapped[str] = mapped_column(VARCHAR(300), nullable=False)
    default_batch_size: Mapped[int] = mapped_column(Integer, nullable=False)
    higher_is_better_metric: Mapped[bool] = mapped_column(Boolean, default=True)
    eta_knob: Mapped[float] = mapped_column(Float, default=0.5)
    beta_knob: Mapped[Optional[float]] = mapped_column(Float, nullable=True)
    target_metric: Mapped[float] = mapped_column(Float, default=0.5)
    max_epochs: Mapped[int] = mapped_column(Integer, default=100)
    num_pruning_rounds: Mapped[int] = mapped_column(Integer, default=2)
    window_size: Mapped[int] = mapped_column(Integer, default=10)

    max_power: Mapped[float] = mapped_column(Float, nullable=False)
    number_of_gpus: Mapped[int] = mapped_column(Integer, nullable=False)
    gpu_model: Mapped[str] = mapped_column(VARCHAR(length=30), nullable=False)

    mab_prior_mean: Mapped[float] = mapped_column(Float, default=0.0)
    mab_prior_precision: Mapped[float] = mapped_column(Float, default=0.0)
    mab_num_explorations: Mapped[int] = mapped_column(Integer, default=2)
    mab_seed: Mapped[Optional[int]] = mapped_column(Integer, nullable=True)

    mab_random_generator_state: Mapped[Optional[str]] = mapped_column(
        VARCHAR(length=10000), nullable=True
    )
    exp_default_batch_size: Mapped[int] = mapped_column(Integer, nullable=False)

    stage: Mapped[Stage] = mapped_column(Enum(Stage), default=Stage.Pruning)
    min_cost: Mapped[Optional[float]] = mapped_column(Float, nullable=True)
    min_cost_batch_size: Mapped[int] = mapped_column(Integer, nullable=False)

    batch_sizes: Mapped[list["BatchSizeTable"]] = relationship(
        order_by="BatchSizeTable.batch_size.asc()",
        back_populates="job",
        # always fetch batch size(int) whenever we fetch the job.
        # https://docs.sqlalchemy.org/en/14/orm/loading_relationships.html#relationship-loading-techniques
        lazy="joined",
        # Delete all children if the job gets deleted.
        # https://docs.sqlalchemy.org/en/20/orm/cascades.html
        cascade="all, delete-orphan",
    )


class BatchSizeTable(Base):
    """Batch size states table schema. Represents one batch size of a job.

    (job_id, batch_size) as a pk, and have three states(exploration, measurement, GaussianTs arm state) as fk.
    For explorations and measurements, one-to-many relationship. For arm_state, one-to-(zero or one) relationship.
    """

    __tablename__ = "BatchSize"

    job_id: Mapped[str] = mapped_column(
        ForeignKey(
            "Job.job_id",
            ondelete="CASCADE",
        ),
        primary_key=True,
    )
    batch_size: Mapped[int] = mapped_column(Integer, primary_key=True)

    trials: Mapped[list["TrialTable"]] = relationship(
        back_populates="batch_size_state", cascade="all, delete-orphan"
    )

    arm_state: Mapped[Optional["GaussianTsArmStateTable"]] = relationship(
        back_populates="batch_size_state",  # populates GaussianTsArmState->BatchSize
        # https://stackoverflow.com/questions/39869793/when-do-i-need-to-use-sqlalchemy-back-populates
        cascade="all, delete-orphan",
    )

    job: Mapped["JobTable"] = relationship(back_populates="batch_sizes")


class GaussianTsArmStateTable(Base):
    """Gaussian arm state schema. Represents a gaussian thompson arm states of a batch size.

    Refer to [`GaussianTsArmState`][zeus.optimizer.batch_size.server.batch_size_state.models.GaussianTsArmState] for attributes.
    """

    __tablename__ = "GaussianTsArmState"

    job_id: Mapped[str] = mapped_column(VARCHAR(300), primary_key=True)
    batch_size: Mapped[int] = mapped_column(Integer, primary_key=True)  # arm

    param_mean: Mapped[float] = mapped_column(Float, default=0.0)
    param_precision: Mapped[float] = mapped_column(Float, default=0.0)
    reward_precision: Mapped[float] = mapped_column(Float, default=0.0)
    num_observations: Mapped[int] = mapped_column(Integer, default=0)

    batch_size_state: Mapped["BatchSizeTable"] = relationship(
        back_populates="arm_state"
    )

    __table_args__ = (
        ForeignKeyConstraint(
            [job_id, batch_size],
            [BatchSizeTable.job_id, BatchSizeTable.batch_size],
            ondelete="CASCADE",
        ),
    )


class TrialType(enum.Enum):
    """Type of trial.

    Exploration is a trial done during Pruning stage.
    Concurrent is a trial done as a concurrent job submission.
    MAB is a trial done during the MAB stage.
    """

    Exploration = "Exploration"
    Concurrent = "Concurrent"
    MAB = "MAB"


class TrialStatus(enum.Enum):
    """Status of trial.

    Dispatched means this trial is issued.
    Succeded means trial ended without error.
    Failed means trial ended with error.
    """

    Dispatched = "Dispatched"
    Succeeded = "Succeeded"
    Failed = "Failed"


class TrialTable(Base):
    """Represents each trial of training.

    Refer to [`Trial`][zeus.optimizer.batch_size.server.batch_size_state.models.Trial] for attributes.
    """

    __tablename__ = "Trial"

    job_id: Mapped[str] = mapped_column(VARCHAR(300), primary_key=True, nullable=False)
    batch_size: Mapped[int] = mapped_column(Integer, nullable=False)
    trial_number: Mapped[int] = mapped_column(Integer, primary_key=True, nullable=False)
    start_timestamp: Mapped[datetime] = mapped_column(DateTime, nullable=False)
    type: Mapped[TrialType] = mapped_column(Enum(TrialType), nullable=False)
    status: Mapped[TrialStatus] = mapped_column(Enum(TrialStatus), nullable=False)

    end_timestamp: Mapped[Optional[datetime]] = mapped_column(DateTime, nullable=True)
    time: Mapped[Optional[float]] = mapped_column(Float, nullable=True)
    energy: Mapped[Optional[float]] = mapped_column(Float, nullable=True)
    converged: Mapped[Optional[bool]] = mapped_column(Boolean, nullable=True)

    batch_size_state: Mapped["BatchSizeTable"] = relationship(back_populates="trials")

    __table_args__ = (
        ForeignKeyConstraint(
            [job_id, batch_size],
            [BatchSizeTable.job_id, BatchSizeTable.batch_size],
            ondelete="CASCADE",
        ),
    )



"""The server guides the `PipelineFrequencyOptimizer` with frequency plans.

The server is agnostic to the training framework the
[`PipelineFrequencyOptimizer`][zeus.optimizer.pipeline_frequency.optimizer.PipelineFrequencyOptimizer]
is integrated with. A server is useful because large model training is
typically distributed, and we still need one place to coordinate the
frequency plans. Later, the server will be extended to support complete
online profiling and optimization.
"""



"""Example script for running the Zeus trace-driven simulator."""

from __future__ import annotations

import argparse
from pprint import pprint
from typing import Literal

import pandas as pd

from zeus._legacy.job import Job
from zeus._legacy.policy.optimizer import (
    JITPowerLimitOptimizer,
    PruningGTSBatchSizeOptimizer,
)
from zeus._legacy.simulate import Simulator, HistoryEntry


def parse_args() -> argparse.Namespace:
    """Parse commandline arguments."""
    parser = argparse.ArgumentParser()

    parser.add_argument("--dataset", default="librispeech")
    parser.add_argument("--model", default="deepspeech2")
    parser.add_argument("--optimizer", default="adamw")
    parser.add_argument("--target_metric", type=float, default=40.0)
    parser.add_argument("--max_epochs", type=int, default=16)
    parser.add_argument("--b_0", type=int, default=192)
    parser.add_argument(
        "--gpu", default="v100", choices=["a40", "v100", "p100", "rtx6000"]
    )
    parser.add_argument("--eta_knob", type=float, default=0.5)
    parser.add_argument("--beta_knob", type=float, default=2.0)
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument(
        "--num_recurrence",
        type=int,
        default=None,
        help="If None, 2*|B|*|P| will be used as in the paper.",
    )

    return parser.parse_args()


def read_trace(
    gpu: Literal["a40", "v100", "p100", "rtx6000"]
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Read the train and power trace files as Pandas DataFrames."""
    train_df = pd.DataFrame(pd.read_csv("../../trace/summary_train.csv"))
    power_df = pd.DataFrame(pd.read_csv(f"../../trace/summary_power_{gpu}.csv"))
    return train_df, power_df


def run_simulator(
    job: Job,
    gpu: Literal["a40", "v100", "p100", "rtx6000"],
    eta_knob: float,
    beta_knob: float,
    num_recurrence: int | None,
    seed: int = 1,
) -> list[HistoryEntry]:
    """Run the simulator on the given job."""
    # Read in the train and power traces.
    train_df, power_df = read_trace(gpu)

    # Instantiate optimizers.
    plo = JITPowerLimitOptimizer(verbose=True)
    bso = PruningGTSBatchSizeOptimizer(seed=seed, verbose=True)

    # Instantitate the simulator.
    simulator = Simulator(train_df, power_df, bso, plo, seed=seed, verbose=True)

    # Use 2 * |B| * |P| is num_recurrence is None.
    print(num_recurrence)
    if num_recurrence is None:
        job_df = job.filter_df(train_df.merge(power_df, how="inner"))
        num_recurrence = (
            2 * len(job_df.batch_size.unique()) * len(job_df.power_limit.unique())
        )

    # Run the simulator.
    print(num_recurrence)
    return simulator.simulate_one_job(job, num_recurrence, beta_knob, eta_knob)


def main(args: argparse.Namespace) -> None:
    """Run the main routine."""
    # Instantitate the job specification dataclass.
    job = Job(
        args.dataset,
        args.model,
        args.optimizer,
        args.target_metric,
        args.max_epochs,
        args.b_0,
    )

    # Run the simulator.
    history = run_simulator(
        job, args.gpu, args.eta_knob, args.beta_knob, args.num_recurrence, args.seed
    )

    # Print out the list of HistoryEntry's.
    pprint(history)


if __name__ == "__main__":
    main(parse_args())
